

from .VisBeatImports import *
from .EventList import *
import math

DEFAULT_LEAD_TIME = 0;
DEFAULT_TAIL_TIME = 0;

class Warp(AObject):
    """Warp (class): defines how one time signal should be warped to another. Given primarily as source/target events to be matched.
        Attributes:
            source_events: source events
            target_events: target events
    """

    def VBOBJECT_TYPE(self):
        return 'Warp';

    def __init__(self, path=None):
        AObject.__init__(self, path=path);
        if(path):
            self.loadFile();


    @staticmethod
    def FromEvents(source_events, target_events):
        w = Warp();
        sevents = source_events;
        if(isinstance(source_events, EventList)):
            sevents = source_events.events;

        tevents = target_events;
        if(isinstance(target_events, EventList)):
            tevents = target_events.events;

        w.source_events=sevents;
        w.target_events=tevents;
        # w.repeatShorterEvents();
        return w

    @staticmethod
    def FromEventLists(source_eventlist, target_eventlist):
        w = Warp();
        w.source_events = source_eventlist.events;
        w.target_events = target_eventlist.events;
        # w.repeatShorterEvents();
        return w


    def initializeBlank(self):
        AObject.initializeBlank(self);
        self.source_events = [];
        self.target_events = [];
        # self.a_info['WarpType'] = 'Linear';
        self.warp_func = None;  # Warp.LinearInterp;
        # self.warp_func_st = None;
        # self.warp_func_ts = None;


    def getTargetStart(self, lead=None):
        target_start = self.target_events[0].getStartTime();
        if (lead is None):
            lead = min(target_start, DEFAULT_LEAD_TIME);
        return target_start - lead;


    def getTargetEnd(self, lead=None):
        lastind = min(len(self.source_events), len(self.target_events)) - 1;
        return self.target_events[lastind].getStartTime() + DEFAULT_TAIL_TIME;


    def getSourceStart(self):
        source_start = self.source_events[0].getUnrolledStartTime();
        return source_start;


    def getSourceEnd(self):
        lastind = min(len(self.source_events), len(self.target_events)) - 1;
        return self.source_events[lastind].getUnrolledStartTime();


    def getWarpedSourceStart(self, lead=None):
        source_start = self.source_events[0].getUnrolledStartTime();
        if (lead is None):
            lead = min(source_start, DEFAULT_LEAD_TIME);
        return self.warpSourceTime(source_start - lead);


    def getWarpedSourceEnd(self, tail=None):
        last_event = min(len(self.source_events), len(self.target_events)) - 1;
        source_end = self.source_events[last_event].getUnrolledStartTime();
        if (tail is None):
            tail = DEFAULT_TAIL_TIME;
        source_end = source_end + tail;
        return self.warpSourceTime(source_end);


    def setWarpFunc(self, warp_type, **kwargs):
        if (warp_type == 'square'):
            self.warp_func = [Warp.SquareInterp, Warp.SquareInterp];
        elif (warp_type == 'linear'):
            self.warp_func = [Warp.LinearInterp, Warp.LinearInterp];
        elif (warp_type == 'cubic'):
            self.warp_func = [Warp.CubicInterp, Warp.CubicInterp];
        elif (warp_type == 'quad'):
            self.warp_func = [
                Warp.GetEventWarpFunc(self.source_events, self.target_events, Warp.WFunc_Quadratic(), **kwargs),
                Warp.GetEventWarpFunc(self.target_events, self.source_events, Warp.WFunc_Quadratic(), **kwargs)];
        elif (warp_type == 'mouth'):
            self.warp_func = [
                Warp.GetEventWarpFunc(self.source_events, self.target_events, Warp.WFunc_Mouth(**kwargs), **kwargs),
                Warp.GetEventWarpFunc(self.target_events, self.source_events, Warp.WFunc_Mouth(**kwargs), **kwargs)];
        elif (warp_type == 'weight'):
            self.warp_func = [
                Warp.GetEventWarpFunc(self.source_events, self.target_events,
                                      Warp.WFunc_Weight(use_to_weights=None, **kwargs), **kwargs),
                Warp.GetEventWarpFunc(self.target_events, self.source_events,
                                      Warp.WFunc_Weight(use_to_weights=True, **kwargs), **kwargs)];
        elif (warp_type == 'half_accel'):
            self.warp_func = [
                Warp.GetEventWarpFunc(self.source_events, self.target_events, Warp.WFunc_P(p=0.5),
                                      **kwargs),
                Warp.GetEventWarpFunc(self.target_events, self.source_events, Warp.WFunc_P(p=0.5),
                                      **kwargs)];
        elif (warp_type == 'p'):
            p = kwargs.get('p');
            self.warp_func = [
                Warp.GetEventWarpFunc(self.source_events, self.target_events, Warp.WFunc_P(p=p),
                                      **kwargs),
                Warp.GetEventWarpFunc(self.target_events, self.source_events, Warp.WFunc_P(p=p),
                                      **kwargs)];
        elif (warp_type == 'target_time_source_fraction'):
            self.warp_func = [
                Warp.GetEventWarpFunc(self.source_events, self.target_events,
                                      Warp.WFunc_targettime_sourcefraction(**kwargs),
                                      **kwargs),
                Warp.GetEventWarpFunc(self.target_events, self.source_events,
                                      Warp.WFunc_targettime_sourcefraction(**kwargs),
                                      **kwargs)];
        elif (warp_type == 'target_source_fractions'):
            self.warp_func = [
                Warp.GetEventWarpFunc(self.source_events, self.target_events,
                                      Warp.WFunc_target_source_fractions(**kwargs),
                                      **kwargs),
                Warp.GetEventWarpFunc(self.target_events, self.source_events,
                                      Warp.WFunc_target_source_fractions(**kwargs),
                                      **kwargs)];
        elif (warp_type is not None):
            self.warp_func = [warp_type, warp_type];

        self.setInfo('WarpType', warp_type);
        return;


    def warpSourceTime(self, t):
        return self.warp_func[0](t, a_events=self.source_events, b_events=self.target_events);


    def warpSourceTimes(self, t):
        tw = t.copy();
        for a in range(len(t)):
            tw[a] = self.warpSourceTime(t[a]);
        return tw;


    def warpTargetTime(self, t):
        return self.warp_func[1](t, a_events=self.target_events, b_events=self.source_events);


    def warpTargetTimes(self, t):
        tw = t.copy();
        for a in range(len(t)):
            tw[a] = self.warpTargetTime(t[a]);
        return tw;


    def plot(self, xlim=None, sampling_rate=None, new_figure=None, render_control_points=True, render_labels=True,
             time_range=None, full_source_range=None, **kwargs):
        if (sampling_rate is None):
            sampling_rate = 30;

        source_duration = self.getSourceEnd() - self.getSourceStart();
        old_frame_time = truediv(1.0, sampling_rate);
        target_start = self.getTargetStart();
        target_end = self.getTargetEnd();



        # if(xlim is not None):
        #     target_start=xlim[0]

        target_duration = target_end - target_start;

        new_n_samples = target_duration * sampling_rate;
        target_start_times = np.linspace(target_start, target_end, num=new_n_samples, endpoint=False);

        unwarped_target_times = [];
        for st in target_start_times:
            unwarped_target_times.append(self.warpTargetTime(st));

        if (new_figure):
            fig = plt.figure();

        unwarped_target_times = np.array(unwarped_target_times);
        # unwarped_target_times = unwarped_target_times-unwarped_target_times[0]+target_start;
        plt.plot(target_start_times, unwarped_target_times, '-');

        if (render_control_points):
            lastind = min(len(self.source_events), len(self.target_events)) - 1;
            targeteventtimes = Event.ToStartTimes(self.target_events[:lastind]);
            sourceeventtimes = Event.ToStartTimes(self.source_events[:lastind]);
            plt.plot(targeteventtimes, sourceeventtimes, 'o', label='Control Points');
        if (xlim is not None):
            xrng = [xlim[0] + target_start, xlim[1] + target_start];
            ylim = [self.warpTargetTime(xrng[0]), self.warpTargetTime(xrng[1])];
            plt.xlim(xrng);
            plt.ylim(ylim);

        if (time_range is not None):
            plt.xlim(time_range);

        if (render_labels):
            plt.ylabel('Source Time');
            plt.xlabel('Target Time');

        plt.title('Warp Curve')

        if (new_figure is not None):
            return fig;


    def plotImage(self, xlim=None, sampling_rate=None):
        if (sampling_rate is None):
            sampling_rate = 30;
        target_start = self.getTargetStart();
        target_end = self.getTargetEnd() + 10;
        target_duration = target_end - target_start;

        new_n_samples = target_duration * sampling_rate;
        target_start_times = np.linspace(target_start, target_end, num=new_n_samples, endpoint=True);

        unwarped_target_times = [];
        for st in target_start_times:
            unwarped_target_times.append(self.warpTargetTime(st));

        unwarped_target_times = np.array(unwarped_target_times);
        pim = Image.PlotImage(signal=unwarped_target_times, show_axis=True, xvals=target_start_times,
                              sampling_rate=sampling_rate, events=self.target_events, xlime=[0, 100], ylims=[0, 110]);
        return pim;


    def repeatShorterEvents(self, endpoints=False):
        n_events = max(len(self.source_events), len(self.target_events));
        self.source_events = Event.RepeatToLength(self.source_events, n=n_events, endpoints=endpoints);
        self.target_events = Event.RepeatToLength(self.target_events, n=n_events, endpoints=endpoints);


    @staticmethod
    def FromEvents(source_events, target_events):
        w = Warp();
        w.source_events = source_events;
        w.target_events = target_events;
        # w.repeatShorterEvents();
        return w


    @staticmethod
    def LinearInterp(t, a_events, b_events):
        n_events = min(len(a_events), len(b_events));
        next_a_event_index = n_events;
        for s in range(n_events):
            if (t < a_events[s].start):
                next_a_event_index = s;
                break;

        prev_a_event_time = 0;
        prev_b_event_time = 0;
        if (next_a_event_index > 0):
            prev_a_event_time = a_events[next_a_event_index - 1].start;
            prev_b_event_time = b_events[next_a_event_index - 1].start;

        next_a_event_time = a_events[n_events - 1].start;
        next_b_event_time = b_events[n_events - 1].start;
        if (next_a_event_index < n_events):
            next_a_event_time = a_events[next_a_event_index].start;
            next_b_event_time = b_events[next_a_event_index].start;

        a_event_gap = next_a_event_time - prev_a_event_time;
        # b_event_gap = next_b_event_time - prev_b_event_time;
        t_progress = t - prev_a_event_time;

        # take care of past-the-end case, by simply letting time proceed normally past the last event
        if (a_event_gap == 0):
            return prev_b_event_time + t_progress;

        next_weight = t_progress / (1.0 * a_event_gap);
        return (next_weight * next_b_event_time) + ((1.0 - next_weight) * prev_b_event_time);


    # additional_points = [];
    # for i in range(-10, 11):
    #     additional_points.append([0.1 * i, 0.51])
    @staticmethod
    def plotWarpMethodTest(warp_type, additional_points=None, **kwargs):
        sev = [];
        tev = [];

        pts = [[1.0, 1.0],
               [-1.0, 1.0],
               [0.5, 1.0],
               [1.0, 0.5],
               [-1.0, 0.5],
               [0.5, 0.5],
               [1.0, 0.3],
               [-1.0, 0.3],
               [0.5, 0.3]];

        if (additional_points is not None):
            pts = pts + additional_points;

        currentt = [0.0, 0.0];
        sev.append(Event(start=0.0));
        tev.append(Event(start=0.0));
        for p in pts:
            currentt[0] = currentt[0] + p[0];
            currentt[1] = currentt[1] + p[1];
            sev.append(Event(start=currentt[0]));
            tev.append(Event(start=currentt[1]));

        # warp_type = 'target_time_source_fraction';
        # other_params['acceleration_target_time'] = 0.5;
        # other_params['acceleration_source_fraction'] = 0.75;
        warpf = Warp.FromEvents(sev, tev);
        warpf.setWarpFunc(warp_type, **kwargs);

        warpf.plot()


    @staticmethod
    def SquareInterp(t, a_events, b_events):
        n_events = min(len(a_events), len(b_events));
        next_a_event_index = n_events;
        for s in range(n_events):
            if (t < a_events[s].start):
                next_a_event_index = s;
                break;

        prev_a_event_time = 0;
        prev_b_event_time = 0;
        if (next_a_event_index > 0):
            prev_a_event_time = a_events[next_a_event_index - 1].start;
            prev_b_event_time = b_events[next_a_event_index - 1].start;

        next_a_event_time = a_events[n_events - 1].start;
        next_b_event_time = b_events[n_events - 1].start;
        if (next_a_event_index < n_events):
            next_a_event_time = a_events[next_a_event_index].start;
            next_b_event_time = b_events[next_a_event_index].start;

        a_event_gap = next_a_event_time - prev_a_event_time;
        # b_event_gap = next_b_event_time - prev_b_event_time;
        t_progress = t - prev_a_event_time;

        # take care of past-the-end case, by simply letting time proceed normally past the last event
        if (a_event_gap == 0):
            return prev_b_event_time + t_progress;

        progress_frac = t_progress / (1.0 * a_event_gap);
        next_weight = math.pow(progress_frac, 2);
        # accel = 3;
        # bweight = math.pow(progress_frac,accel);
        # aweight = math.pow(1.0-progress_frac,accel);
        # sumweight = aweight+bweight;
        # next_weight=bweight/sumweight;
        return (next_weight * next_b_event_time) + ((1.0 - next_weight) * prev_b_event_time);


    @staticmethod
    def GetEventWarpFunc(from_events, to_events, f, lead_time=None, **kwargs):
        start_cap_time = min(from_events[0].start, to_events[0].start);
        if (lead_time is not None):
            start_cap_time = min(start_cap_time, lead_time);
        n_events = min(len(from_events), len(to_events));
        # f_events = Event.GetUnrolledList(from_events[:n_events], assert_on_folds=True);
        f_events = Event.GetUnrolledList(from_events[:n_events]);
        t_events = Event.GetUnrolledList(to_events[:n_events]);

        def rfunc(t, **kwargs):
            next_f_event_index = n_events - 1;
            for e in range(n_events):
                if (t < f_events[e].unrolled_start):
                    next_f_event_index = e;
                    break;

            if (next_f_event_index == 0):
                from_cap_event = Event(start=f_events[0].start - start_cap_time, weight=0, type='startcap');
                to_cap_event = Event(start=t_events[0].start - start_cap_time, weight=0, type='startcap')
                return f(t, f_neighbors=[from_cap_event, f_events[0]], t_neighbors=[to_cap_event, t_events[0]]);

            else:
                return f(t,
                         f_neighbors=[f_events[next_f_event_index - 1], f_events[next_f_event_index]],
                         t_neighbors=[t_events[next_f_event_index - 1], t_events[next_f_event_index]]);

        return rfunc;


    @staticmethod
    def WFunc_Quadratic():
        def rfunc(t, f_neighbors, t_neighbors, **kwargs):
            from_times = Event.ToStartTimes(f_neighbors);
            to_times = Event.ToStartTimes(t_neighbors);
            from_event_gap = from_times[1] - from_times[0];
            t_progress = t - from_times[0];
            # avoid divide by 0
            if (from_event_gap == 0):
                AWARN("Event Gap was 0! Check Warp.py")
                return prev_to_event_time + t_progress;
            progress_frac = truediv(t_progress, from_event_gap);

            next_weight = math.pow(progress_frac, 2);

            return (next_weight * to_times[1]) + ((1.0 - next_weight) * to_times[0]);

        return rfunc;


    @staticmethod
    def WFunc_Weight(use_to_weights=None, **kwargs):
        print("USING WEIGHT-BASED WARP");

        def rfunc(t, f_neighbors, t_neighbors):
            from_times = Event.ToStartTimes(f_neighbors);
            to_times = Event.ToStartTimes(t_neighbors);
            from_event_gap = from_times[1] - from_times[0];
            to_event_gap = to_times[1] - to_times[0];
            t_progress = t - from_times[0];
            # avoid divide by 0
            if (from_event_gap == 0):
                AWARN("Event Gap was 0! Check Warp.py")
                return prev_to_event_time + t_progress;
            progress_frac = truediv(t_progress, from_event_gap);

            if (use_to_weights):
                weight = t_neighbors[1].weight;
            else:
                weight = f_neighbors[1].weight;

            p = 1.0 - weight;
            # a = truediv(1.0, (1.0+p*p-2*p)); # if a=b
            a = 1.0 - np.power((1.0 - p), 2.0);  # b=1
            if (progress_frac < p):
                next_weight = a * progress_frac;
            else:
                next_weight = (a * progress_frac) + np.power((progress_frac - p), 2.0);
            return (next_weight * to_times[1]) + ((1.0 - next_weight) * to_times[0]);

        return rfunc;


    @staticmethod
    def WFunc_P(p=None, **kwargs):
        if (p is None):
            p = 0.5;
        print(("USING P WARP with P={}".format(p)));

        def rfunc(t, f_neighbors, t_neighbors):
            from_times = Event.ToStartTimes(f_neighbors);
            to_times = Event.ToStartTimes(t_neighbors);
            from_event_gap = from_times[1] - from_times[0];
            to_event_gap = to_times[1] - to_times[0];
            t_progress = t - from_times[0];
            # avoid divide by 0
            if (from_event_gap == 0):
                AWARN("Event Gap was 0! Check Warp.py")
                return prev_to_event_time + t_progress;
            progress_frac = truediv(t_progress, from_event_gap);
            # a = truediv(1.0, (1.0+p*p-2*p)); # if a=b
            a = 1.0 - np.power((1.0 - p), 2.0);  # b=1
            if (progress_frac < p):
                next_weight = a * progress_frac;
            else:
                next_weight = (a * progress_frac) + np.power((progress_frac - p), 2.0);
            return (next_weight * to_times[1]) + ((1.0 - next_weight) * to_times[0]);

        return rfunc;


    @staticmethod
    def WFunc_Mouth(p_acceleration_time=0.1, **kwargs):
        def rfunc(t, f_neighbors, t_neighbors):
            from_times = Event.ToStartTimes(f_neighbors);
            to_times = Event.ToStartTimes(t_neighbors);
            from_event_gap = from_times[1] - from_times[0];
            to_event_gap = to_times[1] - to_times[0];
            t_progress = t - from_times[0];
            # avoid divide by 0
            if (from_event_gap == 0):
                AWARN("Event Gap was 0! Check Warp.py")
                return prev_to_event_time + t_progress;
            progress_frac = truediv(t_progress, from_event_gap);

            if (f_neighbors[1].type == 'mouth_open' or t_neighbors[1].type == 'mouth_open'):
                p = 1.0 - truediv(p_acceleration_time, to_event_gap);
                if (p < 0):
                    next_weight = math.pow(progress_frac, 2);
                else:
                    # a = truediv(1.0, (1.0+p*p-2*p)); # if a=b
                    a = 1.0 - np.power((1.0 - p), 2.0);  # b=1
                    if (progress_frac < p):
                        next_weight = a * progress_frac;
                    else:
                        next_weight = (a * progress_frac) + np.power((progress_frac - p), 2.0);
            else:
                next_weight = progress_frac;

            return (next_weight * to_times[1]) + ((1.0 - next_weight) * to_times[0]);

        return rfunc;


    @staticmethod
    def WFunc_targettime_sourcefraction(acceleration_target_time=0.1, acceleration_source_fraction=0.8, **kwargs):
        """
        This assumes that you are mapping from the target to the source, as is the most common use case.
        :param acceleration_target_time: amount of from time to spend accelerating
        :param acceleration_source_fraction: fraction of source to accelerate through
        :return:
        """
        lin_source_fraction = 1.0 - acceleration_source_fraction;

        def rfunc(t, f_neighbors, t_neighbors):
            from_times = Event.ToStartTimes(f_neighbors);
            to_times = Event.ToStartTimes(t_neighbors);
            from_event_gap = from_times[1] - from_times[0];
            to_event_gap = to_times[1] - to_times[0];
            t_progress = t - from_times[0];
            # avoid divide by 0
            if (from_event_gap == 0):
                AWARN("Event Gap was 0! Check Warp.py")
                return to_times[
                           0] + t_progress;  # is this right? doesnt seem to come up, not sure what behavior should be looking back on it...

            progress_frac = truediv(t_progress, from_event_gap);

            time_left = from_event_gap - t_progress;

            p = 1.0 - truediv(acceleration_target_time, from_event_gap);
            p2 = p * p;
            q = 1.0 - acceleration_source_fraction;

            if (acceleration_target_time >= from_event_gap or q > (p2)):
                next_weight = math.pow(progress_frac, 2);
            elif (time_left >= acceleration_target_time):
                lin_t_progress_frac = truediv(t_progress, from_event_gap - acceleration_target_time);
                next_weight = lin_t_progress_frac * lin_source_fraction;
            else:
                pdnom = (p2 - 2.0 * p + 1.0)

                # I just used matlab to symbolic inverse matrix of equations, then simplified by hand
                a = ((1.0 - q) / pdnom) + (q / (p * (p - 1)));
                b = ((2 * p * (q - 1)) / pdnom) - ((q * (p + 1)) / (p2 - p))
                # c = ((p2 - (q * (2 * p - 1))) / pdnom) + (q / (p - 1));
                c = 1 - a - b;

                next_weight = a * progress_frac * progress_frac + b * progress_frac + c;
            return (next_weight * to_times[1]) + ((1.0 - next_weight) * to_times[0]);

        return rfunc;


    @staticmethod
    def WFunc_target_source_fractions(acceleration_target_fraction=0.8, acceleration_source_fraction=0.9, **kwargs):
        """
        This assumes that you are mapping from the target to the source, as is the most common use case.
        :param acceleration_target_fraction: fraction of target to spend accelerating
        :param acceleration_source_fraction: fraction of source to accelerate through
        :return:
        """
        lin_source_fraction = 1.0 - acceleration_source_fraction;



        def rfunc(t, f_neighbors, t_neighbors):
            print(acceleration_target_fraction)
            print(acceleration_source_fraction)
            from_times = Event.ToStartTimes(f_neighbors);
            to_times = Event.ToStartTimes(t_neighbors);
            from_event_gap = from_times[1] - from_times[0];
            to_event_gap = to_times[1] - to_times[0];
            t_progress = t - from_times[0];
            # avoid divide by 0
            if (from_event_gap == 0):
                AWARN("Event Gap was 0! Check Warp.py")
                return to_times[
                           0] + t_progress;  # is this right? doesnt seem to come up, not sure what behavior should be looking back on it...

            progress_frac = truediv(t_progress, from_event_gap);

            time_left = from_event_gap - t_progress;

            p = 1.0 - acceleration_target_fraction;
            p2 = p * p;
            q = 1.0 - acceleration_source_fraction;

            if (acceleration_target_fraction >= 1 or q > (p2)):
                next_weight = math.pow(progress_frac, 2);
            elif (progress_frac <= p):
                lin_t_progress_frac = truediv(t_progress, p * from_event_gap);
                next_weight = lin_t_progress_frac * lin_source_fraction;
            else:
                pdnom = (p2 - 2.0 * p + 1.0)

                # I just used matlab to symbolic inverse matrix of equations, then simplified by hand
                a = ((1.0 - q) / pdnom) + (q / (p * (p - 1)));
                b = ((2 * p * (q - 1)) / pdnom) - ((q * (p + 1)) / (p2 - p))
                # c = ((p2 - (q * (2 * p - 1))) / pdnom) + (q / (p - 1));
                c = 1 - a - b;

                next_weight = a * progress_frac * progress_frac + b * progress_frac + c;
            return (next_weight * to_times[1]) + ((1.0 - next_weight) * to_times[0]);

        return rfunc;


    @staticmethod
    def WFunc_AB(const_factor, quad_factor, **kwargs):
        def rfunc(t, f_neighbors, t_neighbors):
            from_times = Event.ToStartTimes(f_neighbors);
            to_times = Event.ToStartTimes(t_neighbors);
            from_event_gap = from_times[1] - from_times[0];
            t_progress = t - from_times[0];
            # avoid divide by 0
            if (from_event_gap == 0):
                AWARN("Event Gap was 0! Check Warp.py")
                return prev_to_event_time + t_progress;
            progress_frac = truediv(t_progress, from_event_gap);

            next_weight = math.pow(progress_frac, 2);

            return (next_weight * to_times[1]) + ((1.0 - next_weight) * to_times[0]);


    @staticmethod
    def ABWarp():
        next_from_event_time = from_events[n_events - 1].start;
        next_to_event_time = to_events[n_events - 1].start;
        if (next_from_event_index < n_events):
            next_from_event_time = from_events[next_from_event_index].start;
            next_to_event_time = to_events[next_to_event_index].start;

        from_event_gap = next_from_event_time - prev_from_event_time;
        # b_event_gap = next_b_event_time - prev_b_event_time;
        t_progress = t - prev_from_event_time;

        # take care of past-the-end case, by simply letting time proceed normally past the last event
        if (from_event_gap == 0):
            return prev_to_event_time + t_progress;

        progress_frac = t_progress / (1.0 * from_event_gap);
        next_weight = math.pow(progress_frac, 2);
        return (next_weight * next_to_event_time) + ((1.0 - next_weight) * prev_to_event_time);


    @staticmethod
    def CubicInterp(t, a_events, b_events):
        # def CubicInterp(a_events, b_events, t):
        f = Warp.CubicInterpFunc(a_events=a_events, b_events=b_events);
        return f(t);


    @staticmethod
    def LinearInterpFunc(a_events, b_events):
        if (a_events[0].start > 0):
            ae = np.concatenate((np.asarray([0]), Event.ToStartTimes(a_events)));
        else:
            ae = Event.ToStartTimes(a_events);
        be = Event.ToStartTimes(b_events);
        if (len(be) > len(ae)):
            be = be[:len(ae)];
        elif (len(ae) > len(be)):
            ae = ae[:len(be)];
        return sp.interpolate.interp1d(ae, be, 'linear', bounds_error=False, fill_value='extrapolate');


    @staticmethod
    def CubicInterpFunc(a_events, b_events):
        event_times = Event.ToStartTimes(a_events);
        wevent_times = Event.ToStartTimes(b_events);
        minlen = min(len(event_times), len(wevent_times));
        event_times = event_times[:minlen];
        wevent_times = wevent_times[:minlen];
        splinecap = np.arange(0, 1, 0.25);

        spline_times = np.concatenate((splinecap + event_times[0] - 1, event_times));
        wspline_times = np.concatenate((splinecap + wevent_times[0] - 1, wevent_times));

        spline_times = np.append(spline_times, splinecap + 0.5 + spline_times[-1]);
        wspline_times = np.append(wspline_times, splinecap + 0.5 + wspline_times[-1]);

        # get spline to translate initial times to warped times
        splineX = sp.interpolate.interp1d(spline_times, wspline_times, kind='cubic', bounds_error=False,
                                          fill_value='extrapolate');
        return splineX;



from .VisBeatDefines import *
from .AImports import *
from .AObject import AObject
import numpy as np
import scipy as sp

import os
import imageio

import matplotlib
try:
    import matplotlib.pyplot as plt
except ImportError as e:
    AWARN("matplotlib problem... if you are using conda try installing with 'conda install matplotlib'")
    matplotlib.use('agg');
    import matplotlib.pyplot as plt
import matplotlib.style as ms

import io
import base64
import math
from operator import truediv
import time
import shutil
from time import gmtime, strftime, localtime
import librosa;
from ._mediafiles import GetVBMarkPath

def local_time_string():
    return strftime("%Y-%m-%d_%H:%M:%S", localtime());


def VBWARN(message):
    print(message)
    #
    # def send_warnings_to_print_red(message, category, filename, lineno):
    #     print(colored('{} WARNING! file: {} Line:{}\n{}'.format(category, filename, lineno, message), 'red'))
    # old_showwarning = warnings.showwarning
    # warnings.showwarning = send_warnings_to_print_red;

VB_MACHINE_ID = None;
# if(VB_MACHINE_ID):
    # matplotlib.use('PS');

ISNOTEBOOK = False;
if(runningInNotebook()):
    ISNOTEBOOK = True;
    import IPython;
    import IPython.display
    ms.use('seaborn-muted')
    if(not runningInSpyder()):
        get_ipython().magic('matplotlib inline')
        from IPython.lib import kernel
        connection_file_path = kernel.get_connection_file()
        connection_file = os.path.basename(connection_file_path)
        kernel_id = connection_file.split('-', 1)[1].split('.')[0]
        # print("Kernel ID:\n{}".format(kernel_id));
    from IPython.display import HTML
    VBIPY = IPython;
    #%matplotlib inline
# else:
    # matplotlib.use('PS');


def vb_get_ipython():
    return VBIPY;

def clipping_params(clip_bins=30, clip_fraction=0.95):
    return dict(clip_bins=clip_bins, clip_fraction=clip_fraction);

def get_hist_clipped(signal, clip_bins=30, clip_fraction=0.95):
    holdshape = signal.shape[:];
    sigrav = signal.copy().ravel();
    sigh, sigb = np.histogram(sigrav, bins=clip_bins);
    maxbini=np.argmax(sigh);
    totalmass = np.sum(sigh);
    total_included = truediv(sigh[maxbini],totalmass);
    prevbini = maxbini;
    nextbini=maxbini;
    bins_included = 1;
    icounter=0;
    while(total_included<clip_fraction and bins_included<clip_bins):
        if((prevbini>=0 and sigh[prevbini]==0) and (nextbini<(clip_bins) and sigh[nextbini]==0)):
            prevbini=prevbini-1;
            nextbini=nextbini+1;
        else:
            if((prevbini>=0 and sigh[prevbini]>0) or nextbini>=clip_bins):
                prevbini=prevbini-1;
            if((nextbini<(clip_bins) and sigh[nextbini]>0) or prevbini<0):
                nextbini=nextbini+1;
        included_segment=sigh[max(prevbini,0):min(nextbini+1, clip_bins)];
        total_included = truediv(np.sum(included_segment), totalmass);
        bins_included=len(included_segment);
        icounter+=1;
    clipsig = np.clip(a=sigrav, a_min=sigb[max(prevbini,0)], a_max=sigb[min(nextbini,clip_bins)]);
    clipsig.shape=signal.shape
    return clipsig;

def np_scale_to_range(data, value_range=None):
    if(value_range is None):
        value_range = [0.0,255.0];
    d = data.copy().ravel();
    currentmin=np.min(d);
    currentmax=np.max(d);
    currentscale = currentmax-currentmin;
    if(currentscale==0):
        VBWARN("CANNOT SCALE CONSTANT ARRAY TO RANGE")
        return;
    divscale = truediv(1.0,currentscale);
    newrange=(value_range[1]-value_range[0]);
    d = (d*divscale)*newrange;
    newmin = (currentmin*divscale)*newrange;
    d = d-newmin+value_range[0]
    d.shape=data.shape;
    return d





# %load VisBeat/MidiParse.py
import mido
import numpy as np
import os
# from TimeSignal1D import *
from .Audio import *
from .Event import *

class VBMIDITrack(object):
    def __init__(self, track, ticks_per_beat):
        self.mido_track = track;
        self.name = track.name;
        # print('Track {}: {}'.format(i, track.name));
        self.tempo = next((msg.tempo for msg in track if msg.type == 'set_tempo'), None);
        self.ticks_per_beat = ticks_per_beat;
        # self._get_note_on_times();
        self.note_on_times = None;


    def getBPM(self):
        return mido.tempo2bpm(self.tempo);

    def getNoteOnTimes(self, include_negative=None):
        if(self.note_on_times is None or (include_negative)):
            self._get_note_on_times(include_negative = include_negative);
        return self.note_on_times;


    def _get_note_on_times(self, include_negative = None):
        """
        Gets the starting time of each note.
        """
        note_times = []
        current_time = 0;
        for msg in self.mido_track:
            if not msg.is_meta:
                delta_time = mido.tick2second(msg.time, self.ticks_per_beat, self.tempo)
                current_time += delta_time
                if msg.type == 'note_on' and msg.velocity > 0:
                    if (include_negative or current_time>=0):
                        note_times.append(current_time);


        self.note_on_times = np.array(note_times);


    def get_note_durations(self, track_num):
        note_durations = []
        currently_played_notes = {}
        current_time = 0
        for msg in self.mido_track:
            if not msg.is_meta:
                delta_time = mido.tick2second(msg.time, self.ticks_per_beat, self.tempo)
                current_time += delta_time
                if msg.type == 'note_on' and msg.velocity > 0 and msg.note not in currently_played_notes:
                    currently_played_notes[msg.note] = current_time
                    #if len(currently_played_notes) > 1:
                    #    print "Number of played notes have reached: ", len(currently_played_notes)
                elif msg.type == 'note_off' or (msg.type == 'note_on' and msg.velocity == 0):
                    duration = current_time - currently_played_notes[msg.note]
                    note_durations.append((currently_played_notes[msg.note], duration))
                    currently_played_notes.pop(msg.note)
        assert not bool(currently_played_notes), "Finished looping through all messages. There should not be any notes playing at this point."
        #print "before sorting: ", note_durations
        note_durations.sort(key=lambda x : x[0])
        #print "after sorting: ", note_durations
        return np.array([ x[1] for x in note_durations ])

    def get_mouth_events(self):
        # note_times = []
        current_time = 0;
        number_on = 0;
        events = [];
        open_buffer = 0.1;
        close_buffer = 0.1
        last_type = 0;
        events.append(Event(start=0, type='mouth_close', weight=1));
        for msg in self.mido_track:
            if not msg.is_meta:
                delta_time = mido.tick2second(msg.time, self.ticks_per_beat, self.tempo)
                current_time += delta_time
                last_time = events[-1].start;
                passed = current_time-last_time;

                if msg.type == 'note_on' and msg.velocity > 0:
                    if(last_type==0):
                        if(passed>open_buffer):
                            events.append(Event(start=current_time-open_buffer, type='mouth_closed', weight=0));
                    events.append(Event(start=current_time, type='mouth_open', weight = truediv(msg.velocity,127)));
                    number_on = number_on + 1;
                    last_type=1;
                if(msg.type == 'note_off' or (msg.type=='note_on' and msg.velocity == 0)):
                    if (last_type == 1):
                        if (passed > close_buffer):
                            events.append(Event(start=current_time - close_buffer, type='mouth_opened', weight=0));
                    events.append(Event(start=current_time, type='mouth_close', weight = truediv(msg.velocity,127)));
                    number_on = number_on-1;
                    last_type=0;
                    # assert(number_on>-1);
                    # if(number_on==0):
        return events;
        # self.note_on_times = np.array(note_times);

    def getNoteOnTimesAsAudio(self, sampling_rate = None, note_sound=None, n_seconds=None):
        assert(n_seconds is not None), "must provide n seconds"
        if(sampling_rate is None):
            sampling_rate = 16000;
        if(note_sound is None):
            note_sound = Audio.getPing();
        s = Audio.Silence(n_seconds=n_seconds, sampling_rate=sampling_rate, name=self.name)
        s = s.getWithSoundAdded(note_sound, self.getNoteOnTimes());
        return s;



class VBMIDI(TimeSignal1D):
    def __init__(self, path=None):
        TimeSignal1D.__init__(self, path=path);
        self.midi_file = mido.MidiFile(path)
        self.tracks = [];
        for i, track in enumerate(self.midi_file.tracks):
            self.tracks.append(VBMIDITrack(track, self.midi_file.ticks_per_beat));

    def getNoteOnTimes(self, include_negative=None):
        return self.tracks[-1].getNoteOnTimes(include_negative=include_negative);

    def getMouthEvents(self):
        return self.tracks[-1].get_mouth_events();

    def getNoteOnTimesAsAudio(self, sampling_rate=None, note_sound=None):
        s = self.tracks[-1].getNoteOnTimesAsAudio(sampling_rate = sampling_rate, note_sound = note_sound, n_seconds = self.midi_file.length);
        if (s.name is None):
            s.name = self.getInfo('file_name')
        return s;


#AImports
from . import ADefines as defines
import os
import os.path
import errno
import json
import pickle as pickle
import glob
import subprocess
from operator import truediv
import shutil
import time
from time import gmtime, strftime, localtime
import random
from . import fileui



try:
    from termcolor import colored
    def AWARN(message):
        if (defines.AD_DEBUG):
            print((colored(message, 'red')))
    def AINFORM(message):
        if(defines.AD_DEBUG):
            print((colored(message, 'blue')))
except ImportError:
    if (defines.AD_DEBUG):
        print("You do not have termcolor installed (pip install termcolor). AWARN will just show as plain print statements when defines.AD_DEBUG==True...")
    def AWARN(message):
        if (defines.AD_DEBUG):
            print(message);
    def AINFORM(message):
        if (defines.AD_DEBUG):
            print(message);

def local_time_string():
    return strftime("%Y-%m-%d_%H:%M:%S", localtime());

def get_temp_file_path(final_file_path="TEMP", temp_dir_path = None):
    pparts = os.path.split(final_file_path);
    destfolder = pparts[0]+os.sep;
    tempdir = temp_dir_path;
    if(tempdir is None):
        tempdir='.';
    destfolder=pathstring(tempdir+os.sep);
    tempname = 'TEMP_'+pparts[1];
    temptry = 0;
    while(os.path.isfile(destfolder+tempname)):
        temptry=temptry+1;
        tempname = 'TEMP{}_'.format(temptry)+pparts[1];
    return pathstring(destfolder+tempname);


def runningInNotebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter

def getshellname():
    try:
        shell = get_ipython().__class__.__name__
        return shell;
    except NameError:
        return False      # Probably standard Python interpreter


def runningInSpyder():
    return 'SpyderKernel' in str(get_ipython().kernel.__class__);

def pickleToPath(d, path):
    # assert(False)
    print("pickling")
    f = open(path, 'wb');
    pickle.dump(d, f, protocol=2);
    f.close();
    return True;

def unpickleFromPath(path):
    f=open(path, 'rb');
    d=pickle.load(f);
    f.close();
    return d;

def make_sure_path_exists(path):
    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise

def make_sure_dir_exists(path):
    pparts = os.path.split(path);
    destfolder = pparts[0]+os.sep;
    try:
        os.makedirs(destfolder)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise

def safe_file_name(input_string):
    return ''.join([i if ord(i) < 128 else '_' for i in input_string]);

def pathstring(path):
    return path.replace(os.sep+os.sep, os.sep);

def is_interactive():
    import __main__ as main
    return not hasattr(main, '__file__')

def printOb(obj):
    for attr in dir(obj):
        print("#### Obj.%s = %s\n" % (attr, getattr(obj, attr)))

def pathstring(path):
    return path.replace(os.sep+os.sep, os.sep);

def get_prepended_name_file_path(original_file_path, string_to_prepend):
    pparts = os.path.split(original_file_path);
    destfolder = pparts[0]+os.sep;
    pname = string_to_prepend+pparts[1];
    return pathstring(destfolder+pname);

def safe_file_name(input_string):
    return ''.join([i if ord(i) < 128 else '_' for i in input_string]);

def change_extension(input_path, new_ext):
    nameparts = os.path.splitext(input_path);
    return nameparts[0]+new_ext;

def printDictionary(obj):
    if type(obj) == dict:
        for k, v in list(obj.items()):
            if hasattr(v, '__iter__'):
                print(k)
                printDictionary(v)
            else:
                print('%s : %s' % (k, v))
    elif type(obj) == list:
        for v in obj:
            if hasattr(v, '__iter__'):
                printDictionary(v)
            else:
                print(v)
    else:
        print(obj)

def spotgt_shift_bit_length(x):
    #smallest power of two greater than
    return 1<<(x-1).bit_length()

def get_file_name_from_path(pth):
    return os.path.split(pth)[1];

def get_dir_from_path(pth):
    return (os.path.split(pth)[0]+os.sep);

def get_file_names_from_paths(pths):
    r = [];
    for p in pths:
        r.append(get_file_name_from_path(p));
    return r;

def writeDictionaryToJSON(d, json_path=None):
    if(json_path):
        with open(json_path, 'w') as outfile:
            json.dump(d, outfile, sort_keys = True, indent = 4, ensure_ascii=False);


def vtt_to_srt(fileContents):
    replacement = re.sub(r'([\d]+)\.([\d]+)', r'\1,\2', fileContents)
    replacement = re.sub(r'WEBVTT\n\n', '', replacement)
    replacement = re.sub(r'^\d+\n', '', replacement)
    replacement = re.sub(r'\n\d+\n', '\n', replacement)
    return replacement



#VideoVersion#from VisBeatImports import *
from .AObject import *
from .TimeSignal import *
from .Audio import *
from .Warp import *
from .Image import *
import moviepy.editor as mpy
from moviepy.audio.AudioClip import AudioArrayClip as MPYAudioArrayClip

import sys

import math
from operator import truediv


def MPYWriteVideoFile(mpyclip, filename, **kwargs):
    temp_audio_filename = get_temp_file_path(final_file_path='TEMP_'+filename+'.m4a', temp_dir_path=Video.VIDEO_TEMP_DIR);
    return mpyclip.write_videofile(filename=filename, temp_audiofile= temp_audio_filename, audio_codec='aac', **kwargs);



class Video(TimeSignal):
    """Video (class): A video, and a bunch of convenience functions to go with it.
        Attributes:
    """

    VIDEO_TEMP_DIR = './'
    FEATURE_FUNCS = TimeSignal.FEATURE_FUNCS.copy();

    def AOBJECT_TYPE(self):
        return 'Video';

    def __init__(self, path=None, name=None, num_frames_total=None):
        TimeSignal.__init__(self, path=path);
        if(name):
            self.name = name;
        if(path):
            self.loadFile(num_frames_total=num_frames_total);

    def initializeBlank(self):
        TimeSignal.initializeBlank(self);#YES KEEP
        self.name = None;
        self.sampling_rate = None;
        self.audio=None;
        self.reader = None;
        self.writer = None;
        self.num_frames_total=None;
        self.reshape=None;
        self.meta_data = None;
        self.sampling_rate = None;
        self.source = None;

        # _gui is not saved
        self._gui = None;

    def _getFrameRate(self):
        return self.sampling_rate;

    # _gui is from TimeSignal
    # <editor-fold desc="Property: 'gui'">
    @property
    def gui(self):
        return self._getGui();

    def _getGui(self):
        return self._gui;

    @gui.setter
    def gui(self, value):
        self._setGui(value);

    def _setGui(self, value):
        self._gui = value;
    # </editor-fold>


    def getVersionInfo(self):
        """This is the info to be saved in asset version dictionary"""
        d={};
        d['name']=self.name;
        d['sampling_rate']=self.sampling_rate;
        d['num_frames_total']=self.num_frames_total;
        d['duration']=self.getDuration();
        d['start_time']=self.getStartTime();
        d['end_time']=self.getEndTime();
        d['meta_data']=self.meta_data;
        return d;

    def getName(self):
        if(self.name is None):
            return self.getInfo('file_name');
        else:
            return self.name;

    def getTempDir(self):
        if(self.source is not None):
            return self.source.getDir('temp');
        else:
            return Video.VIDEO_TEMP_DIR;

    def getStringForHTMLStreamingBase64(self):
        svideo = io.open(self.getPath(), 'r+b').read()
        encoded = base64.b64encode(svideo)
        return "data:video/mp4;base64,{0}".format(encoded.decode('ascii'));


    def n_frames(self):
        if(not self.num_frames_total):
            self.num_frames_total = self.calcNumFramesTotal();
        return self.num_frames_total;

    def getDuration(self):
        return truediv(self.n_frames(), self.sampling_rate);

    def getStartTime(self):
        return 0;

    def getEndTime(self):
        return self.getDuration();

    def getMPYClip(self, get_audio=True):
        return mpy.VideoFileClip(self.getPath(), audio=get_audio);

    def getAudio(self):
        return self.audio;

    def loadFile(self, file_path=None, num_frames_total=None):
        if (file_path):
            self.setPath(file_path=file_path);
        if ('file_path' in self.a_info):
            self.reader = imageio.get_reader(self.a_info['file_path'], 'ffmpeg');
            self.meta_data = self.reader.get_meta_data();
            self.sampling_rate = self.meta_data['fps'];
            if (num_frames_total is not None):
                self.num_frames_total = num_frames_total;
            else:
                self.num_frames_total = self.calcNumFramesTotal();

            try:
                self.audio = Audio(self.a_info['file_path']);
                self.audio.name =self.name;
            #except RuntimeError:
            except Exception:
                print(("Issue loading audio for {}".format(self.a_info['file_path'].encode('utf-8'))));
                self.audio = Audio(sampling_rate=16000);
                self.audio.x = np.zeros(int(np.ceil(self.audio.sampling_rate*self.getDuration())));

    def openVideoWriter(self, output_file_path, fps=None):
        if('outputs' not in self.a_info):
            self.a_info['outputs'] = [];
        out_fps = fps;
        if(not out_fps):
            out_fps=self.sampling_rate;
        make_sure_dir_exists(output_file_path);
        self.writer = imageio.get_writer(output_file_path, 'ffmpeg', macro_block_size = None, fps = out_fps);
        self.a_info['outputs'].append(output_file_path);

    def closeVideoWriter(self):
        self.writer.close();
        self.writer = None;

    def getFrameShape(self):
        fs=self.getInfo('frame_shape');
        if(fs is not None):
            return fs;
        else:
            self.setInfo(label='frame_shape', value=self.reader.get_data(0).shape);
            return self.getInfo('frame_shape');

    def calcNumFramesTotal(self):
        #assert(False)
        print(("Calculating frames for {}...".format(self.name)))
        valid_frames = 0
        example_frame = self.reader.get_data(0);
        self.setInfo(label='frame_shape',value=example_frame.shape);
        
        # https://stackoverflow.com/questions/54778001/how-to-to-tackle-overflowerror-cannot-convert-float-infinity-to-integer
        #for i in range(1, self.reader.get_length()):
        for i in range(1, self.reader.count_frames()):
            try:
                self.reader.get_data(i);
            except imageio.core.format.CannotReadFrameError as e:
                break
            valid_frames += 1
        print("Done.")
        return valid_frames

    def readFrameBasic(self, i):
        """You should basically never call this"""
        fi=i;
        if(fi<0):
            fi=0;
        if(fi>(self.num_frames_total-1)):
            fi=(self.num_frames_total-1);
        return np.asarray(self.reader.get_data(int(fi)));

    def getFrame(self, f):
        return self.readFrameBasic(round(f));

    def getFrameFromTime(self, t):
        f = t*self.sampling_rate;
        return self.getFrame(f=f);

    def getFrameLinearInterp(self, f):
        if(isinstance(f,int) or f.is_integer()):
            return self.readFrameBasic(int(f));
        prev_frame = self.readFrameBasic(math.floor(f));
        next_frame = self.readFrameBasic(math.ceil(f));
        frame_progress = f-np.floor(f);
        rframe =  (next_frame*frame_progress)+(prev_frame*(1.0-frame_progress));
        return rframe.astype(prev_frame.dtype);

    def writeFrame(self, img):
        if self.writer.closed:
            print('ERROR: Vid writer object is closed.')
        else:
            self.writer.append_data(img.astype(np.uint8))

    def play(self):
        if(ISNOTEBOOK):
            print("Playing video:")
            video = io.open(self.getPath(), 'r+b').read()
            encoded = base64.b64encode(video)
            vidhtml=HTML(data='''<video alt="test" controls>
                <source src="data:video/mp4;base64,{0}" type="video/mp4" />
             </video>'''.format(encoded.decode('ascii')));
            IPython.display.display(vidhtml);
            # return vidhtml;
        else:
             print("HOW TO PLAY VIDEO? NOT A NOTEBOOK.")

    def show(self):
        self.play();

    def write(self, output_path, output_sampling_rate=None):
        assert output_path, "MUST PROVIDE OUTPUT PATH FOR VIDEO"

        sampling_rate = output_sampling_rate;
        if(not sampling_rate):
            sampling_rate=self.sampling_rate;
        tempfilepath = get_temp_file_path(final_file_path=output_path, temp_dir_path=self.getTempDir());
        self.openVideoWriter(output_file_path=tempfilepath, fps=output_sampling_rate);

        duration = self.getDuration();
        nsamples = sampling_rate*duration;
        old_frame_time = truediv(1.0,self.sampling_rate);
        frame_start_times = np.linspace(0,self.getDuration(),num=nsamples,endpoint=False);
        frame_index_floats = frame_start_times*self.sampling_rate;

        start_timer=time.time();
        last_timer=start_timer;
        fcounter=0;
        for nf in range(len(frame_index_floats)):
            self.writeFrame(self.getFrame(frame_index_floats[nf]));
            fcounter+=1;
            if(not (fcounter%50)):
                if((time.time()-last_timer)>10):
                    last_timer=time.time();
                    print(("{}%% done after {} seconds...".format(100.0*truediv(fcounter,len(frame_index_floats)), last_timer-start_timer)));


        self.closeVideoWriter();
        rvid = Video.CreateFromVideoAndAudio(video_path=tempfilepath, audio_object=self.audio, output_path=output_path);
        os.remove(tempfilepath);
        return rvid;

    def VideoClip(self, start=None, end=None, name=None):
        from . import VideoClip
        if(start is None):
            start = 0;
        if(end is None):
            end = self.getDuration();
        clip = VideoClip.VideoClip(video=self, start=start, end=end);
        if(name):
            clip.name=name;
        return clip;

    def getImageFromFrame(self, i):
        rimage = Image(data=self.getFrame(i));
        rimage.setInfo(label='parent_video', value=self.getInfo('file_path'));
        rimage.setInfo(label='frame_number',value=i);
        return rimage;

    def getImageFromTime(self, t):
        rimage = Image(data=self.getFrameFromTime(t));
        rimage.setInfo(label='parent_video', value=self.getInfo('file_path'));
        rimage.setInfo(label='frame_time', value=t);
        return rimage;


    # def getFrameShape(self):
    #     return self.getImageFromFrame(0).getShape();

    def writeResolutionCopyFFMPEG(self, path, max_height=None):
        if(max_height is None):
            max_height=self.getFrameShape()[0];
        mpc = self.getMPYClip();
        clip_resized = mpc.resize(height=max_height); # make the height 360px ( According to moviePy documenation The width is then computed so that the width/height ratio is conserved.)
        if((clip_resized.size[0]%2)>0):
            clip_resized=clip_resized.crop(x1=0,width=clip_resized.size[0]-1);

        # clip_resized.write_videofile(path);
        MPYWriteVideoFile(clip_resized, path);

        rvid = Video(path);
        return rvid;

    def writeFFMPEG(self, output_path):
        mpc = self.getMPYClip();
        # mpc.write_videofile(output_path, preset='fast', codec='mpeg4');
        MPYWriteVideoFile(mpc, output_path, preset='fast', codec='mpeg4')
        rvid = Video(output_path);
        return rvid;

    def writeResolutionCopy(self, path, max_height=None, reshape=None, input_sampling_rate_factor = None, output_sampling_rate=None):
        print((self.getPath()))
        if(input_sampling_rate_factor is not None):
            original_sampling_rate = self.sampling_rate;
            self.sampling_rate=input_sampling_rate_factor*self.sampling_rate;
            original_audio_sampling_rate = self.audio.sampling_rate;
            self.audio.sampling_rate=self.audio.sampling_rate*input_sampling_rate_factor;
        output_path = path;
        inshape = self.getFrameShape();
        if(reshape==True):
            imshape = [inshape[1],inshape[0],inshape[2]];
        outshape = None;
        if(max_height and (max_height<inshape[0])):
            outshape = self.getFrameShape();
            out_w_over_height = truediv(inshape[1],inshape[0]);
            outshape[0]=(max_height);
            outshape[1]=max_height*out_w_over_height;
            if((math.floor(outshape[1])%2)==0):
                outshape[1]=math.floor(outshape[1])
            elif((math.ceil(outshape[1])%2)==0):
                outshape[1]=math.ceil(outshape[1])
            if((outshape[1]%2)!=0):
                outshape[1]=outshape[1]-1;
        sampling_rate = output_sampling_rate;
        if(not sampling_rate):
            sampling_rate=self.sampling_rate;

        tempfilepath = get_temp_file_path(final_file_path=output_path, temp_dir_path=self.getTempDir());
        self.openVideoWriter(output_file_path=tempfilepath, fps=output_sampling_rate);

        duration = self.getDuration();
        nsamples = sampling_rate*duration;
        old_frame_time = truediv(1.0,self.sampling_rate);
        frame_start_times = np.linspace(0,self.getDuration(),num=nsamples,endpoint=False);
        frame_index_floats = frame_start_times*self.sampling_rate;

        start_timer=time.time();
        last_timer=start_timer;
        fcounter=0;
        for nf in range(len(frame_index_floats)):
            fim = self.getImageFromFrame(frame_index_floats[nf]);
            if(reshape==True):
                fim.data = np.reshape(fim.data, (fim.data.shape[1],fim.data.shape[0],fim.data.shape[2]));
            if(outshape is not None):
                fim = fim.getScaled(shape=outshape);
            self.writeFrame(fim.data);

            fcounter+=1;
            if(not (fcounter%50)):
                if((time.time()-last_timer)>10):
                    last_timer=time.time();
                    print(("{}%% done after {} seconds...".format(100.0*truediv(fcounter,len(frame_index_floats)), last_timer-start_timer)));
        self.closeVideoWriter();
        rvid = Video.CreateFromVideoAndAudio(video_path=tempfilepath, audio_object=self.audio, output_path=output_path);
        os.remove(tempfilepath);
        if(input_sampling_rate_factor is not None):
            #return original_sampling_rate
            self.sampling_rate=original_sampling_rate;
            self.audio.sampling_rate = original_audio_sampling_rate;
        return rvid;

    def writeWarped(self, output_path, warp, output_sampling_rate=None, output_audio=None, bitrate=None, vbmark = True, max_time = None, **kwargs):
        sampling_rate = output_sampling_rate;
        if (not sampling_rate):
            sampling_rate = self.sampling_rate;

        duration = self.getDuration();
        old_frame_time = truediv(1.0, self.sampling_rate);

        target_start = warp.getTargetStart();
        target_end = warp.getTargetEnd();

        target_duration = target_end - target_start;

        if(max_time is not None and target_duration>max_time):
            target_duration = max_time;
            target_end = target_start+max_time;

        print((
            "target start: {}\ntarget end: {}\ntarget duration: {}".format(target_start, target_end, target_duration)));

        new_n_samples = target_duration * sampling_rate;
        target_start_times = np.linspace(target_start, target_end, num=new_n_samples, endpoint=False);
        unwarped_target_times = [];
        for st in target_start_times:
            unwarped_target_times.append(warp.warpTargetTime(st));
        frame_index_floats = np.true_divide(np.array(unwarped_target_times), old_frame_time);
        tempfilepath = get_temp_file_path(final_file_path=output_path, temp_dir_path=Video.VIDEO_TEMP_DIR);
        self.openVideoWriter(output_file_path=tempfilepath, fps=output_sampling_rate, **kwargs);
        start_timer = time.time();
        last_timer = start_timer;
        fcounter = 0;
        if(vbmark):
            vbmarker = self.getImageFromFrame(0)._vbmarker();

        for nf in range(len(frame_index_floats)):
            try:
                nwfr = self.getFrame(frame_index_floats[nf]);
                if(vbmark):
                    nwfrm = Image(data=nwfr);
                    nwfrm._splatAtPixCoord(**vbmarker);
                    nwfr = nwfrm._pixels_uint;
                self.writeFrame(nwfr);
            except ValueError:
                print(("VALUE ERROR ON WRITEFRAME {}".format(frame_index_floats[nf])));
                self.writeFrame(self.getFrame(math.floor(frame_index_floats[nf])));
            fcounter += 1;
            if (not (fcounter % 50)):
                if ((time.time() - last_timer) > 10):
                    last_timer = time.time();
                    print(("{}%% done after {} seconds...".format(100.0 * truediv(fcounter, len(frame_index_floats)),
                                                                 last_timer - start_timer)));
        self.closeVideoWriter();

        silent_warped_vid = Video(tempfilepath);
        if (not output_audio):
            output_audio = self.getAudio();

        cropped_output_audio = output_audio.AudioClip(start=target_start, end=target_end);

        if ((self.getInfo('max_height') is None) and (bitrate is None)):
            use_bitrate = "20000k";
            print(('Using bitrate of {}'.format(use_bitrate)));
            rvid = Video.CreateFromVideoAndAudioObjects(video=silent_warped_vid, audio=cropped_output_audio,
                                                        output_path=output_path, bitrate=use_bitrate);
        elif (bitrate is 'regular'):
            rvid = Video.CreateFromVideoAndAudioObjects(video=silent_warped_vid, audio=cropped_output_audio,
                                                        output_path=output_path);
        else:
            rvid = Video.CreateFromVideoAndAudioObjects(video=silent_warped_vid, audio=cropped_output_audio,
                                                        output_path=output_path, bitrate=bitrate);

        os.remove(tempfilepath);
        rvid.setInfo(label='warp_used', value=warp);
        return rvid;



    def getVersionLabel(self):
        return self.getInfo('version_label');

    def getWarpsDir(self):
        if(self.source is None):
            return self.getTempDir();
        version_label = self.getVersionLabel()
        return self.source.getWarpsDir(version_label=version_label);

    def getWithBeginningCroppedToAudio(self, target):
        if(isinstance(target, Audio)):
            B=target;
        else:
            B=target.getAudio();

        A = self.getAudio();

        AB = A.getShiftAmountTo(B);
        BA = B.getShiftAmountTo(A);

        if(AB<BA):
            return self.VideoClip(name=self.name+'clipped_to_'+B.name);
        else:
            return self.VideoClip(start=BA, end=None, name=self.name+'clipped_to_'+B.name);

    def getWarped(self, target,
                  source_events=None, target_events=None,
                  warp_type=None,
                  max_time = None,
                  source_time_range = None, target_time_range=None,
                  output_sampling_rate=None, output_path=None,
                  force_recompute=None, name_tag = None, vbmark = True, **kwargs):
        """

        :param target_audio:
        :param source_eventlist:
        :param target_eventlist:
        :param warp_type:
        :param source_time_range:
        :param target_time_range:
        :param output_sampling_rate:
        :param output_path:
        :param force_recompute:
        :param name_tag:
        :param kwargs:
        :return:
        """


        if(isinstance(target, Video)):
            target_audio = target.audio;
        else:
            target_audio = target;

        if(source_events is None):
            source_events = self.audio.getBeatEvents();
        if(target_events is None):
            target_events = target_audio.getBeatEvents();

        warp = Warp.FromEvents(source_events, target_events);

        if(warp_type is None):
            warp_type = 'quad';
        warp.setWarpFunc(warp_type, **kwargs);

        if(output_sampling_rate is None):
            output_sampling_rate = self.sampling_rate;

        if(hasattr(target, 'name')):
            warpname=self.name+'_TO_'+target.name;
        else:
            warpname = self.name+ '_TO_TARGETAUDIO' + '_' + str(target.getInfo('name'));

        if(name_tag is not None):
            warpname = warpname+name_tag;

        warpname = warpname+'.mp4';

        if(output_path is None):
            output_path = self.getWarpsDir()+os.sep+warpname;
            make_sure_dir_exists(output_path);

        if(not os.path.isfile(output_path) or force_recompute):
            rvid = self.writeWarped(output_path=output_path, warp=warp, output_sampling_rate=output_sampling_rate, output_audio=target_audio, vbmark = vbmark, max_time = max_time);
        else:
            rvid = Video(path=output_path, name=warpname);
        rvid.setFeature(name='warp_used', value=warp);
        return rvid;


    def getWithSoundsOnEvents(self, events, output_path=None, name_tag = None, mute_original=True,
                              event_sound=None, force_recompute=None, **kwargs):
        """

        :param events: Can be EventList, a list of events, or a list of numbers representing the times of the events.
        :param output_path: if None, will try to get from Video.source
        :param name_tag: what to label this output with in its name.
        :param mute_original: whether to leave the original sound from the video in the output.
        :param event_sound: Sound to play at each event.
        :param force_recompute:
        :param kwargs:
        :return:
        """

        assert(events is not None), 'must provide events or list of times to put sounds'

        output_name = self.getName()+'_eventsounds';
        if (name_tag):
            output_name = output_name + '_' + name_tag;
        if(output_path is None):
            if(self.source is not None):
                output_dir = self.source.getDirForVersion(version_label=self.getInfo('version_label'), version_group='sound_on_events');
            else:
                output_dir = self.getTempDir();
            output_path=os.path.join(output_dir, output_name+'.mp4');
            make_sure_dir_exists(output_dir);
        if(not os.path.isfile(output_path) or force_recompute):
            if(isinstance(events, EventList)):
                etimes = events.toStartTimes();
            elif(isinstance(events[0], Event)):
                etimes = Event.ToStartTimes(events);
            else:
                etimes = events;
            new_audio = self.getAudio().getWithSoundAdded(sound=event_sound, add_times=etimes,
                                                          mute_original=mute_original, **kwargs);
            audio_sig = new_audio.getSignal();
            reshapex = audio_sig.reshape(len(audio_sig), 1);
            reshapex = np.concatenate((reshapex, reshapex), axis=1);
            audio_clip = MPYAudioArrayClip(reshapex, fps=new_audio.sampling_rate);  # from a numeric arra
            video_clip = self.getMPYClip();
            video_clip = video_clip.set_audio(audio_clip);
            # video_clip.write_videofile(output_path, codec='libx264', write_logfile=False);
            MPYWriteVideoFile(video_clip, output_path, codec='libx264', write_logfile=False);

        rvid = Video(path=output_path, name=output_name);
        return rvid;


    @staticmethod
    def CreateFromVideoAndAudio(video_path=None, audio_path=None, video_object=None, audio_object=None,
                                output_path=None, clip_to_video_length=True, return_vid=True, codec='libx264',
                                bitrate=None, **kwargs):
        assert (not (video_path and video_object)), "provided both video path and object to CreateFromVideoAndAudio"
        assert (not (audio_path and audio_object)), "provided both audio path and object to CreateFromVideoAndAudio"
        assert (output_path), "Must provide output path for CreateFromVideoAndAudio"

        if (video_path):
            video_object = Video(video_path);
        if (audio_path):
            audio_object = Audio(path=audio_path);

        output_path = output_path.encode(sys.getfilesystemencoding()).strip();
        make_sure_dir_exists(output_path);

        # audio_sig = audio_object.getSignal();
        audio_sig = audio_object.getStereo();
        audio_sampling_rate = audio_object.getStereoSamplingRate();
        is_stereo = True;

        if (audio_sig is None):
            is_stereo = False;
            audio_sig = audio_object.getSignal();
            audio_sampling_rate = audio_object.sampling_rate;
            n_audio_samples_sig = len(audio_sig);
        else:
            n_audio_samples_sig = audio_sig.shape[1];

        print(("stereo is {}".format(is_stereo)));

        audio_duration = audio_object.getDuration();
        video_duration = video_object.getDuration();

        if (clip_to_video_length):
            n_audio_samples_in_vid = int(math.ceil(video_duration * audio_sampling_rate));

            if (n_audio_samples_in_vid < n_audio_samples_sig):
                if (is_stereo):
                    audio_sig = audio_sig[:, :int(n_audio_samples_in_vid)];
                else:
                    audio_sig = audio_sig[:int(n_audio_samples_in_vid)];
            else:
                if (n_audio_samples_in_vid > n_audio_samples_sig):
                    nreps = math.ceil(truediv(n_audio_samples_in_vid, n_audio_samples_sig));
                    if (is_stereo):
                        audio_sig = np.concatenate(
                            (audio_sig, np.zeros((2, n_audio_samples_in_vid - n_audio_samples_sig))),
                            axis=1);
                    else:
                        audio_sig = np.tile(audio_sig, (int(nreps)));
                        audio_sig = audio_sig[:int(n_audio_samples_in_vid)];

        if (is_stereo):
            # reshapex = np.reshape(audio_sig, (audio_sig.shape[1], audio_sig.shape[0]), order='F');
            reshapex = np.transpose(audio_sig);
            audio_clip = MPYAudioArrayClip(reshapex, fps=audio_sampling_rate);  # from a numeric arra
        else:
            reshapex = audio_sig.reshape(len(audio_sig), 1);
            reshapex = np.concatenate((reshapex, reshapex), axis=1);
            audio_clip = MPYAudioArrayClip(reshapex, fps=audio_sampling_rate);  # from a numeric arra

        video_clip = video_object.getMPYClip();
        video_clip = video_clip.set_audio(audio_clip);
        # video_clip.write_videofile(output_path,codec='libx264', write_logfile=False);
        if (bitrate is None):
            # video_clip.write_videofile(output_path, codec=codec, write_logfile=False);
            MPYWriteVideoFile(video_clip, output_path, codec=codec, write_logfile=False);
        else:
            MPYWriteVideoFile(video_clip, output_path, codec=codec, write_logfile=False, bitrate=bitrate);
            # video_clip.write_videofile(output_path, codec=codec, write_logfile=False, bitrate=bitrate);
        if (return_vid):
            return Video(output_path);
        else:
            return True;

    @staticmethod
    def CreateByStackingVideos(video_objects=None, video_paths=None, output_path=None, audio = None, concatdim = 0, force_recompute=True, **kwargs):
        assert output_path, "MUST PROVIDE OUTPUT PATH FOR VIDEO"
        if(not force_recompute):
            if(os.path.isfile(output_path)):
                return Video(path=output_path);
        matchdim = (concatdim+1)%2;
        vids=[];
        if(video_objects is not None):
            vids=video_objects;

        if(video_paths is not None):
            for vp in video_paths:
                vids.append(Video(path=video_paths));

        output_path=output_path.encode(sys.getfilesystemencoding()).strip();
        make_sure_dir_exists(output_path);
        basevid = vids[0];
        if(audio is None):
            audio = basevid.audio;
        sampling_rate = basevid.sampling_rate;
        tempfilepath = get_temp_file_path(final_file_path=output_path, temp_dir_path=vids[0].getTempDir());
        basevid.openVideoWriter(output_file_path=tempfilepath, fps=sampling_rate);
        duration = basevid.getDuration();
        nsamples = sampling_rate*duration;
        old_frame_time = truediv(1.0,sampling_rate);
        #frame_start_times = np.linspace(self.getStartTime(),self.getEndTime(),num=nsamples,endpoint=False);
        frame_start_times = np.linspace(0,duration,num=nsamples,endpoint=False);
        #frame_index_floats = frame_start_times/old_frame_time;
        frame_index_floats = frame_start_times*sampling_rate;
        for nf in range(len(frame_index_floats)):
            frameind = frame_index_floats[nf];
            newframe = basevid.getFrame(frameind);
            for vn in range(1,len(vids)):
                addpart = vids[vn].getFrame(frameind);
                partsize = np.asarray(addpart.shape)[:];
                cumulsize = np.asarray(newframe.shape)[:];
                if(partsize[matchdim]!=cumulsize[matchdim]):
                    sz = partsize[:];
                    sz[matchdim]=cumulsize[matchdim];
                    addpart = sp.misc.imresize(addpart, size=sz);
                newframe = np.concatenate((newframe, addpart), concatdim);
            basevid.writeFrame(newframe);
        basevid.closeVideoWriter();
        rvid = Video.CreateFromVideoAndAudio(video_path=tempfilepath, audio_object=audio, output_path=output_path, **kwargs);
        os.remove(tempfilepath);
        return rvid;

    @staticmethod
    def CreateFromVideoAndAudioPaths(video_path, audio_path, output_path, return_vid = True, **kwargs):
        return Video.CreateFromVideoAndAudio(video_path=video_path, audio_path=audio_path, output_path=output_path,return_vid=return_vid, **kwargs);

    @staticmethod
    def CreateFromVideoAndAudioObjects(video, audio, output_path, clip_to_video_length=True, return_vid = True, **kwargs):
        return Video.CreateFromVideoAndAudio(video_object=video, audio_object=audio, output_path=output_path, clip_to_video_length=clip_to_video_length, return_vid = return_vid, **kwargs);


    def _getDefaultPeakPickingTimeParams(self, **kwargs):
        single_frame = np.true_divide(1.0, self._getFrameRate());
        time_params = dict(
            pre_max_time=2.0 * single_frame,
            post_max_time=2.0 * single_frame,
            pre_avg_time=5.0 * single_frame,
            post_avg_time=5.0 * single_frame,
            wait_time=2.0 * single_frame,
            delta=0.015,
        )
        time_params.update(kwargs);
        return time_params;

#   ###################

    # FEATURE
    def getFrameIndexes(self, force_recompute=False):
        feature_name = 'frame_indexes';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            params = dict();
            duration = self.getDuration();
            nsamples = self.sampling_rate * duration;
            frame_start_times = np.linspace(0, duration, num=nsamples, endpoint=False);
            value = frame_start_times * self.sampling_rate;
            self.setFeature(name=feature_name, value=value, params=params);
        return self.getFeature(feature_name);


    def getFeaturesList(self):
        return super(Video, self).getFeaturesList()+self.audio.getFeaturesList();

    def getVideoFeaturesList(self):
        return super(Video, self).getFeaturesList();

    def getFeatureFunctionsList(self):
        return super(Video, self).getFeatureFunctionsList()+self.audio.getFeatureFunctionsList();

    def getFeatureSourceType(self, name):
        """returns whether video or audio feature, instead of parent version which returns true or false"""
        vidhas = super(Video, self).hasFeature(name=name);
        if(vidhas):
            return 'video';
        audiohas = self.audio.hasFeature(name=name);
        if(audiohas):
            return 'audio';
        return None;

    def getFeature(self, name, force_recompute=False, **kwargs):
        """Returns None if feature is not already computed, and feature name is not implemented and registered with FEATURE_FUNCS"""
        params = kwargs;
        assert(not kwargs.get('params')), "STILL TRYING TO USE PARAMS INSTEAD OF KWARGS. FIX THIS";
        ftry = super(Video, self).getFeature(name=name, force_recompute=force_recompute, **kwargs);
        if(ftry is not None):
            return ftry;
        else:
            return self.audio.getFeature(name=name, force_recompute=force_recompute, **kwargs);


Video.FEATURE_FUNCS['frame_indexes']=Video.getFrameIndexes;
from . import Video_CV;
if(Video_CV.USING_OPENCV):
    Video.FEATURE_FUNCS.update(Video_CV.FEATURE_FUNCS);
    Video.USING_OPENCV = Video_CV.USING_OPENCV;

    Video.localRhythmicSaliencyFunction = Video_CV.localRhythmicSaliencyFunction
    Video.visualBeatFunction = Video_CV.visualBeatFunction

    Video.cvGetGrayFrame = Video_CV.cvGetGrayFrame;
    Video.getImageFromFrameGray = Video_CV.getImageFromFrameGray;
    Video.plotVisualBeats = Video_CV.plotVisualBeats;
    Video.loadFlowFeatures = Video_CV.loadFlowFeatures;

    Video.getVisualBeats = Video_CV.getVisualBeats
    Video.getLocalRhythmicSaliency = Video_CV.getLocalRhythmicSaliency;
    Video.getDirectogram = Video_CV.getDirectogram;
    Video.getDirectogramPowers = Video_CV.getDirectogramPowers;
    Video.getVisibleImpactEnvelope = Video_CV.getVisibleImpactEnvelope;
    Video.getVisibleImpactEnvelopePowers = Video_CV.getVisibleImpactEnvelopePowers;
    Video.getVisibleImpacts = Video_CV.getVisibleImpacts;
    Video.getVisualBeats = Video_CV.getVisualBeats;
    # Video.getBackwardVisualBeats = Video_CV.getBackwardVisualBeats;
    # Video.getForwardVisualBeats = Video_CV.getForwardVisualBeats;
    Video.getBackwardVisibleImpactEnvelope = Video_CV.getBackwardVisibleImpactEnvelope;
    Video.getBothWayVisibleImpactEnvelope = Video_CV.getBothWayVisibleImpactEnvelope;
    Video.getForwardVisibleImpactEnvelope = Video_CV.getForwardVisibleImpactEnvelope;
    Video.getDirectionalFlux = Video_CV.getDirectionalFlux;
    Video.getVisualTempogram = Video_CV.getVisualTempogram;
    Video.getCutEvents = Video_CV.getCutEvents;
    Video.computeImpactEnvelope = Video_CV.computeImpactEnvelope;
    Video.computeDirectogramPowers = Video_CV.computeDirectogramPowers;
    Video.visualBeatsFromEvents = Video_CV.visualBeatsFromEvents;
    Video.plotVisualBeats = Video_CV.plotVisualBeats;
    Video.plotImpactEnvelope = Video_CV.plotImpactEnvelope;
    Video.plotVisibleImpacts = Video_CV.plotVisibleImpacts;
    Video.getVisualBeatSequences = Video_CV.getVisualBeatSequences;
    Video.printVisualBeatSequences = Video_CV.printVisualBeatSequences;
    Video.getVisualBeatTimes = Video_CV.getVisualBeatTimes;
    Video.findAccidentalDanceSequences = Video_CV.findAccidentalDanceSequences;
    Video.plotEvents = Video_CV.plotEvents;
    # Video. = Video_CV.
    # Video. = Video_CV.
    # Video. = Video_CV.
    # Video. = Video_CV.
    # Video. = Video_CV.
    # Video. = Video_CV.

else:
    AWARN("Was not able to add functions that use OpenCV! Check OpenCV instalation and try again?");

from .vbgui import BeatGUI
if(BeatGUI.VIEWER_INSTALLED):
    def getEvents(self, **kwargs):
        time_params = self._getDefaultPeakPickingTimeParams();
        time_params.update(kwargs)
        vbeats = self.getFeature('visual_beats', force_recompute=True, **time_params);
        return vbeats;
    def getEventList(self, **kwargs):
        events = self.getEvents(**kwargs);
        return EventList(events=events);
    def runBeatGUIOnAudio(self):
        audio = self.getAudio();
        self.gui.run(local_saliency=audio.getLocalRhythmicSaliency(), frame_rate = audio._getFrameRate(), eventlist = audio.getEventList());


    def runGUI(self, local_saliency=None, frame_rate = None, eventlist = 'default', frame_offset=None):
        self.gui.run(local_saliency=local_saliency, frame_rate=frame_rate, eventlist=eventlist, frame_offset=frame_offset);

    Video._getGui = BeatGUI.media_GUI_func;
    Video.runGUI = runGUI;
    Video.getEvents = getEvents;
    Video.getEventList = getEventList;
    Video.runBeatGUIOnAudio = runBeatGUIOnAudio;

else:
    AWARN("BeatGUI not installed");


import os


VB_MEDIA_UTILS_PATH = os.path.abspath(__file__)
VB_MEDIA_UTILS_DIR = os.path.abspath(os.path.dirname(__file__));
MEDIAFILES_DIR = os.path.join(VB_MEDIA_UTILS_DIR, '_assets'+os.sep)

AUDIO_FILES_DIR = os.path.join(MEDIAFILES_DIR, 'audio'+os.sep);
AUDIO_FILES = [];
AUDIO_FILE_PATHS = {};
for filename in os.listdir(AUDIO_FILES_DIR):
    # if(reduce(lambda x,y: x or y, map(lambda ext: filename.lower().endswith(ext), Audio.MEDIA_FILE_EXTENSIONS()))):
    AUDIO_FILES.append(filename);
    AUDIO_FILE_PATHS[filename]=(os.path.join(AUDIO_FILES_DIR, filename));


def GetTestAudioPath(filename):
    return AUDIO_FILE_PATHS[filename];


VIDEO_FILES_DIR = os.path.join(MEDIAFILES_DIR, 'video'+os.sep);
VIDEO_FILES = [];
VIDEO_FILE_PATHS = [];
if(os.path.exists(VIDEO_FILES_DIR)):
    for filename in os.listdir(VIDEO_FILES_DIR):
        VIDEO_FILES.append(filename);
        VIDEO_FILE_PATHS.append(os.path.join(VIDEO_FILES_DIR, filename));



IMAGE_FILES_DIR = os.path.join(MEDIAFILES_DIR, 'images'+os.sep);
IMAGE_FILES = [];
IMAGE_FILE_PATHS = {};
for filename in os.listdir(IMAGE_FILES_DIR):
    IMAGE_FILES.append(filename);
    IMAGE_FILE_PATHS[filename] = (os.path.join(IMAGE_FILES_DIR, filename));

def GetTestImagePath(filename=None):
    if(filename is None):
        filename = "VisBeatWatermark.png"
    return IMAGE_FILE_PATHS[filename];

def GetVBMarkPath():
    return IMAGE_FILE_PATHS["VisBeatWatermark.png"];


from .VisBeatImports import *

class Event(AObject):
    """Event (class): An event in time, either in video or audio
        Attributes:
            start: when the event starts
    """

    DIRECTION_FORWARD = 1;
    DIRECTION_BACKWARD = -1;
    DIRECTION_BOTH = 0;
    _EVENT_PHASE_BASE_RES = 8;

    def AOBJECT_TYPE(self):
        return 'Event';

    def __str__(self):
        return str(self.getAttributeDict());

    def __init__(self, start=None, type=None, weight=None, index=None, is_active=1, unrolled_start = None, direction=0, **kwargs):
        AObject.__init__(self, path=None);
        self.start = start;
        self.type=type;
        self.weight=weight;
        self.index = index;
        self.unrolled_start = unrolled_start;
        self.is_active = is_active;
        self.direction = direction;
        self.a_info.update(kwargs);

    def initializeBlank(self):
        AObject.initializeBlank(self);
        self.start = None;
        self.type = None;
        self.weight = None;
        self.index = None;
        self.unrolled_start = None;
        self.is_active = None;
        self.direction = None;


    def getAttributeDict(self):
        d = dict();
        d['start'] = self.start;
        d['type'] = self.type;
        d['weight'] = self.weight;
        d['index'] = self.index;
        d['unrolled_start'] = self.unrolled_start;
        d['is_active'] = self.is_active;
        d['direction'] = self.direction;
        return d;

    def toDictionary(self):
        d=AObject.toDictionary(self);
        d.update(self.getAttributeDict());
        return d;


    def initAttributesFromDictionary(self, d):
        self.start = d['start'];
        self.type = d.get('type');
        self.weight = d.get('weight');
        self.index = d['index'];
        self.unrolled_start = d.get('unrolled_start');
        self.is_active = d['is_active'];
        if (d.get('direction') is None):
            self.direction = 0;
        else:
            self.direction = d['direction'];

    def initFromDictionary(self, d):
        AObject.initFromDictionary(self, d);
        self.initAttributesFromDictionary(d);

    def clone(self, start=None):
        newe = Event();
        newe.initFromDictionary(self.toDictionary());
        if (start):
            newe.start = start;
        return newe;


    def _getIsSelected(self):
        return self.getInfo('is_selected');
    def _setIsSelected(self, is_selected):
        self.setInfo('is_selected', is_selected);

    def _getPhase(self, phase_resolution=None):
        if (self.getInfo('phase') is None):
            return -1;
        if (phase_resolution is None):
            return self.getInfo('phase');
        else:
            return self.getInfo('phase') * (phase_resolution / Event._EVENT_PHASE_BASE_RES);

    def _setPhase(self, phase, phase_resolution):
        if (phase_resolution is None):
            self.setInfo('phase', phase);
        else:
            self.setInfo('phase', phase * (Event._EVENT_PHASE_BASE_RES / phase_resolution));

    def _getBoundaryType(self):
        return self.getInfo('boundary_type');
    def _setBoundaryType(self, boundary_type):
        self.setInfo('boundary_type', boundary_type);

    @classmethod
    def _FromGUIDict(cls, gd, phase_resolution=None):
        e = cls();
        e.initAttributesFromDictionary(gd);
        e._setPhase(gd['phase'], phase_resolution=phase_resolution);
        e._setBoundaryType(gd.get('boundary_type'));
        e._setIsSelected(gd.get('is_selected'));
        return e;

    @classmethod
    def _FromGUIDicts(cls, gds, type=None):
        events = [];
        for gd in gds:
            ne = cls._FromGUIDict(gd);
            if (type is not None):
                ne.type = type;
            events.append(ne);
        return events;

    def _toGUIDict(self):
        '''
        note that is_active is switched to 0 or 1 for javascript/json
        :return:
        '''
        d = self.getAttributeDict();
        if(self.is_active):
            d['is_active'] = 1;
        else:
            d['is_active'] = 0;
        d['phase'] = self._getPhase();
        d['boundary_type'] = self._getBoundaryType();
        d['is_selected'] = self._getIsSelected();
        return d;

    @staticmethod
    def _ToGUIDicts(events, active=None):
        '''
        convert to dicts for javascript GUI
        :param events:
        :param active: if not none, all of the is_active flags will be set to this
        :return:
        '''
        startind = int(round(time.time() * 1000));
        starts = [];
        for e in range(len(events)):
            de = events[e]._toGUIDict();
            if (active is not None):
                assert (active is 0 or active is 1), "is_active must be 0 or 1";
                de['is_active'] = active;
            if (de.get('index') is None):
                de['index'] = startind + e;
            starts.append(de);
        return starts;


    def getUnrolledStartTime(self):
        if(self.unrolled_start is not None):
            return self.unrolled_start;
        else:
            return self.start;


    def getStartTime(self):
        return self.start;



    def getShifted(self, new_start_time):
        return Event(self.start-new_start_time, type=self.type, weight=self.weight, index=self.index);




    @staticmethod
    def GetUnrolledList(event_list, assert_on_folds=None):
        out_list = [];
        event0 = event_list[0].clone();
        event0.unrolled_start = event0.start;
        event0.index = 0;
        out_list.append(event0);
        for e in range(1,len(event_list)):
            newe = event_list[e].clone();
            if(assert_on_folds):
                assert(newe.start>=out_list[-1].start), 'FOLD (non-monotonic event list) DETECTED WHEN NOT ALLOWED!!!\n see Event.GetUnrolledList'
            newe.unrolled_start = out_list[-1].unrolled_start+np.fabs(newe.start-out_list[-1].start);
            newe.index = e;
            out_list.append(newe);
        return out_list;

    @staticmethod
    def NewFromIndices(event_list, inds):
        out_list = [];
        for ei in inds:
            out_list.append(event_list[ei].clone());
        return out_list;

    @staticmethod
    def RollToNOld(events, n_out, momentum = 0.25):
        inds = [0];
        step = 1;
        n_events = len(events);
        for b in range(1,n_out):
            lastind = inds[-1];
            if ((n_out-b)<(n_events-lastind) or lastind==0):
                inds.append(lastind+1);
                step = 1;
            elif(lastind==(len(events)-1)):
                inds.append(len(events)-2);
                step = -1;
            else:
                foreward_p = 0.5+momentum;
                roll = np.random.rand();
                if(roll<foreward_p):
                    inds.append(lastind+step);
                else:
                    inds.append(lastind-step);
                    step = -step;
        print(inds);
        return Event.GetUnrolledList(Event.NewFromIndices(events, inds));

    @staticmethod
    def UnfoldToN(events, n_out, momentum=0.25):
        # This weird way of implementing is the result of stripping non-visbeat stuff out of the original code...
        return Event.RollToN(events=events, n_out=n_out, momentum=momentum);



    @staticmethod
    def GetDirectedLinks(events):
        links = [{}]
        links[0]['prev_f']=None;
        links[0]['prev_b'] = None;
        lastf = None;
        lastb = None;
        if(events[0].direction<1):
            lastb = 0;
        if(events[0].direction>-1):
            lastf = 0;

        for ei in range(1,len(events)):
            links.append({});
            links[ei]['prev_f']=lastf;
            links[ei]['prev_b']=lastb;

            lastbu = lastb;
            lastfu = lastf;
            if(lastbu is None):
                lastbu = 0;
            if(lastfu is None):
                lastfu = 0;


            if(events[ei].direction<1):
                for lb in range(lastbu,ei):
                    links[lb]['next_b']=ei;
                    # print("ei is {} and lastbu is {}, lastb is {}".format(ei, lastbu, lastb));
                lastb = ei;
            if (events[ei].direction>-1):
                for lf in range(lastfu, ei):
                    links[lf]['next_f'] = ei;
                    # print("ei is {} and lastfu is {}, lastf is {}".format(ei, lastfu, lastf));
                lastf = ei;

        return links;


    @staticmethod
    def RollToN(events, n_out, start_index = 0, momentum=0.1):
        links = Event.GetDirectedLinks(events);
        inds = [start_index];
        step = 1;
        n_events = len(events);
        for b in range(1, n_out):
            lastind = inds[-1];
            foreward_p = 0.5 + momentum;
            roll = np.random.rand();

            if (roll < foreward_p):
                step = step;
            else:
                step = -step;

            if(links[lastind].get('prev_b') is None):
                step = 1;
            if(links[lastind].get('next_f') is None):
                step = -1;

            if(step>0):
                inds.append(links[lastind]['next_f']);
            else:
                inds.append(links[lastind]['prev_b']);
        # print(inds);
        return Event.GetUnrolledList(Event.NewFromIndices(events, inds));



    @staticmethod
    def Clone(event_list):
        out_list = [];
        for ei in event_list:
            out_list.append(ei.clone());
        return out_list;

    @staticmethod
    def SetDirections(event_list, direction):
        for e in range(len(event_list)):
            event_list[e].direction = direction;
        return event_list;



    @classmethod
    def FromStartTimes(cls, starts, type=None):
        events = [];
        for s in starts:
            events.append(cls(start=s, type=type));
        return events;

    @classmethod
    def FromStartsAndWeights(cls, starts, weights, type=None):
        events = [];
        assert(len(starts)==len(weights)), 'Event.FromStartsAndWeights got {} starts and {} weights'.format(len(starts), len(weights));
        for s in range(len(starts)):
            events.append(cls(start=starts[s], weight=weights[s],type=type));
        return events;

    @staticmethod
    def ToStartTimes(events):
        starts = np.zeros(len(events));
        for e in range(len(events)):
            starts[e]=events[e].start;
        return starts;

    @staticmethod
    def ToWeights(events):
        weights = np.zeros(len(events));
        for e in range(len(events)):
            weights[e] = events[e].weight;
        return weights;

    #endpoint is false if events are already placed at beginning and end
    @staticmethod
    def RepeatToLength(events, n, endpoints=False):
        if(n<len(events)):
            return events[:n];
        if(n==len(events)):
            return events;

        if(endpoints):
            print("HAVE NOT IMPLEMENTED ENDPOINTS VERSION OF REPEAT")

        dup_index = 1;
        while(len(events)<n):
            interval = events[dup_index].start-events[dup_index-1].start;
            last_time = events[-1].start;
            newevent = events[dup_index].clone(start=last_time+interval);
            events.append(newevent);
            dup_index=dup_index+1;
        return events;

    @staticmethod
    def Double(events):
        doubled = [];
        for e in range(1,len(events)):
            halfstart = 0.5*(events[e].start+events[e-1].start);
            doubled.append(events[e].clone(start=halfstart));
            doubled.append(events[e]);
        return doubled;

    @staticmethod
    def Half(events, offset=0):
        if(offset is None):
            offset=0;

        halved = [];
        for e in range(1,len(events)):
            if((e%2)==offset):
                halved.append(events[e]);
        return halved;

    @staticmethod
    def SubdivideIntervals(events, extra_samples_per_interval=1):
        newe = [];
        samples = np.linspace(0,1,extra_samples_per_interval+2)[1:-1];
        newe.append(events[0]);
        for e in range(1, len(events)):
            for s in samples:
                newstart = events[e].start*s+events[e-1].start*(1.0-s);
                newe.append(events[e].clone(start=newstart));
            newe.append(events[e].clone())
        return newe;

    @staticmethod
    def Third(events, offset=0):
        if (offset is None):
            offset = 0;
        third = [];
        for e in range(len(events)):
            if (not ((e+offset) % 3)):
                third.append(events[e]);
        return third;

    @staticmethod
    def SubsampleEveryN(events, n, offset=0):
        if (offset is None):
            offset = 0;
        newe = [];
        for e in range(len(events)):
            if (not ((e + offset) % n)):
                newe.append(events[e]);
        return newe;

    # @staticmethod
    # def FromSignalBeats(signal, sampling_rate, event_type=None, **kwargs):
    #     # seg = [0,100]
    #     # xvals = range(seg[1]-seg[0])
    #     xvals = np.arange(len(signal));
    #     xvals = xvals*truediv(1.0, sampling_rate);
    #
    #     vis_tempo, beatinds = librosa.beat.beat_track(y=None, sr=sampling_rate, onset_envelope=signal, hop_length=1, start_bpm=100.0, tightness=80,
    #                                                   trim=True, bpm=None, units='frames');
    #     peaktimes = xvals[beatinds.astype(int)];
    #     peakvals = signal[beatinds.astype(int)];
    #     return Event.FromStartsAndWeights(starts=peaktimes, weights=peakvals, type=event_type);


    @staticmethod
    def FromSignalPeaks(signal, sampling_rate, event_type=None, index_offset=None, **kwargs):
        xvals = np.arange(len(signal));
        xvals = xvals*truediv(1.0, sampling_rate);

        time_params = dict(
            pre_max_time=0.2,
            post_max_time=0.2,
            pre_avg_time=0.2,
            post_avg_time=0.2,
            wait_time=0.1,
        )
        tp_keys = list(time_params.keys());
        time_params.update(kwargs);

        delta = kwargs.get('delta');
        if (delta is None):
            delta = 0.2;

        print(delta)

        for p in tp_keys:
            time_params[p] = int(round(sampling_rate * time_params[p]));

        dparams = dict(
            pre_max=time_params['pre_max_time'],
            post_max=time_params['post_max_time'],
            pre_avg=time_params['pre_avg_time'],
            post_avg=time_params['post_avg_time'],
            wait=time_params['wait_time'],
            delta=delta
        )

        # print(dparams)
        peakinds = librosa.util.peak_pick(x=signal, **dparams);
        # print(peakinds)
        if(not len(peakinds)):
            return [];
        if(index_offset is not None):
            print(('index offset {}'.format(index_offset)))
            for pi in range(len(peakinds)):
                newi = peakinds[pi]+index_offset;
                if(newi>=0 and newi<len(xvals)):
                    print(('{} to {}'.format(peakinds[pi], newi)))
                    peakinds[pi]=newi;


        peaktimes = xvals[peakinds];
        peakvals = signal[peakinds];
        return Event.FromStartsAndWeights(starts=peaktimes, weights=peakvals, type=event_type);


    @staticmethod
    def PlotEventMatches(source_events, target_events, source_in=None, target_in=None):
        t_height = 0.75;
        s_height = 0.25;
        if ((source_in is not None) and (target_in is not None)):
            n_in = min(len(source_in), len(target_in));
            evt = Event.ToStartTimes(target_in[:n_in]);
            evs = Event.ToStartTimes(source_in[:n_in]);
            plt.plot(evt, np.ones(len(evt)) * t_height, 'o');
            plt.plot(evs, np.ones(len(evs)) * s_height, 'o');

        s_et = Event.ToStartTimes(source_events);
        t_et = Event.ToStartTimes(target_events);

        for ev in range(len(target_events)):
            plt.plot([s_et[ev], t_et[ev]], [s_height, t_height], '-');

    @staticmethod
    def GetWithFirstEventAt(events, first_event_time):
        oute = [];
        shift_time = events[0].start-first_event_time;
        for e in events:
            oute.append(e.getShifted(new_start_time=shift_time));
        return oute;

    @staticmethod
    def GetWithStartTimesShifted(events, new_start_time):
        oute = [];
        for e in events:
            oute.append(e.getShifted(new_start_time=new_start_time));
        return oute;

    @staticmethod
    def GetScaled(events, scale):
        oute = [];
        for e in events:
            oute.append(Event(e.start*scale, weight=e.weight, type=e.type, index=e.index));
        return oute;

    @staticmethod
    def GetScaledAndStartingAt(events, scale, starting_at):
        ev = Event.GetScaled(events, scale);
        ev = Event.GetWithFirstEventAt(ev, starting_at);
        return ev;

    @staticmethod
    def ClosestToTargetMatch(source_events, target_events):
        def _closest_e(target_e):
            def sortfunc_e(source_e):
                return math.fabs(source_e[1].start-target_e.start);
            return sortfunc_e;
        def _getinds(s_sort):
            inds = [];
            for s in s_sort:
                inds.append(s[0]);
            return inds;
        def _getevents(s_sorted):
            evnts = [];
            for s in s_sort:
                evnts.append(s[1]);
            return evnts;

        S = [];
        T = [];
        for s in range(len(source_events)):
            S.append([s, source_events[s]]);
        for t in range(len(target_events)):
            T.append([t, target_events[t]]);

        S_sorted = [];
        for t in T:
            S_sorted.append(sorted(S, key=_closest_e(t[1])));

        s_out_inds = [-1]*len(T);
        match_map = {};
        for ti in range(len(S_sorted)):
            # print(S_sorted[ti][0][0]);
            matched_source_event_i = S_sorted[ti][0][0];#sorted lists, first element, index value for index
            existing_match = match_map.get(matched_source_event_i);
            matchdist = math.fabs(source_events[matched_source_event_i].start - target_events[ti].start);
            if(existing_match is not None):
                if(matchdist<existing_match[1]):
                    # print('overrule')
                    s_out_inds[existing_match[0]]=-1;
                    s_out_inds[ti]=matched_source_event_i;
                    # print("ti {}\nmatched {}\n".format(ti, matched_source_event_i));
                    match_map[matched_source_event_i]=[ti, matchdist];
            else:
                match_map[matched_source_event_i] = [ti, matchdist];
                s_out_inds[ti]=matched_source_event_i;

        S_out = [];
        T_out = [];
        # print("s_out_inds: {}".format(s_out_inds))
        for e in range(len(s_out_inds)):
            if(s_out_inds[e]>=0):
                T_out.append(target_events[e]);
                S_out.append(source_events[s_out_inds[e]]);

        return S_out, T_out;

    @staticmethod
    def Sort(event_list, func=None):
        assert(func is None or func=='start' or func=='time'), "have not implemented sort by {}".format(func);
        event_list.sort(key=lambda x: x.start);
        return event_list;

    @staticmethod
    def GetSorted(event_list, func=None):
        clone = Event.Clone(event_list);
        Event.Sort(clone, func=func);
        return clone;

    @staticmethod
    def GetWithTwoWayMerged(event_list, merge_window = 0.1):
        event_list_sorted = Event.GetSorted(event_list);
        new_events = [];
        new_events.append(event_list_sorted[0].clone());
        for ei in range(1,len(event_list_sorted)):
            thise = event_list_sorted[ei];
            if((thise.start-new_events[-1].start)<merge_window):
                if((thise.direction+new_events[-1].direction)==0):
                    new_events[-1].start = 0.5*(thise.start+new_events[-1].start);
                    new_events[-1].direction = 0;
            else:
                new_events.append(thise.clone());
        return new_events


    @staticmethod
    def ApplyIndices(event_list):
        for e in range(len(event_list)):
            event_list[e].index = e;

    @staticmethod
    def PlotSignalAndEvents(signal, sampling_rate, events, time_range = None, ylim=None, **kwargs):
        times = np.arange(len(signal));
        times = times*truediv(1.0,sampling_rate);

        if (kwargs.get('other_events')):
            oet, oev = eventsToTimes(kwargs.get('other_events'), time_range);
            plt.plot(oet, oev, 'x');

        if(events is not None):
            btimes = Event.ToStartTimes(events);
            binds = np.round(btimes*sampling_rate).astype(int);
            bvals = signal[binds];

            mplt = plt.plot(times, signal, '-', btimes, bvals, 'o');
        else:
            mplt = plt.plot(times, signal, '-');

        if(time_range is not None):
            plt.xlim(time_range[0], time_range[1])
        if(ylim is not None):
            plt.ylim(ylim);
        # plt.show()
        return mplt;



from .AImports import *
class AParamDict(object):
    """AParamDict (class): Dictionary that stores values and the parameters used to compute those values. I use this
        class for two things. The first is to store parameters for reproducability. The second is to only recompute
        values when functions are called with different parameters (some of this latter functionality is tied up in
        code that isn't part of the Lite release).

        Attributes:
            data: name -> value, params
    """

    def __init__(self, owner=None, name=None, path=None):
        self.name=name;
        self.data = {};
        self.owner=owner;
        self.modified=False;

    def getEntry(self, name=None, params=None, force_recompute=False):
        d = self.data.get(name);
        if((d is not None) and (not force_recompute)):
            return d;
        return None;

    def setEntry(self, name, d):
        assert(name!='all' and name!='each'),"Entry named '{}' is reserved in AParamDict".format(name);
        self.data[name]=d;

    def removeEntry(self, name, assert_if_absent=True, set_modified = True):
        if(assert_if_absent):
            assert(name in self.data),"Tried to remove entry {} that was not already in {}".format(name, self.__class__);
        popentry = self.data.pop(name, None);
        if(set_modified):
            self.setModified(set_modified);
        return popentry;


    def hasEntry(self, name=None):
        return (self.data.get(name) is not None);

    def getValue(self, name=None, params=None, force_recompute=False):
        d = self.getEntry(name=name, params=params, force_recompute=force_recompute);
        if(d is not None):
            return d.get('value');
        else:
            return None;

    def getParams(self, name=None):
        d = self.data.get(name);
        if(d is not None):
            return d.get('params');
        else:
            return None;

    def setValue(self, name, value=None, params=None, modified=True):
        self.data[name]['value']=value;
        self.data[name]['params']=params;
        self.setEntryModified(name=name, is_modified=modified)

    def saveEntry(self, name, path, force=False):
        """Save one entry to one file."""
        if(self.hasEntry(name=name) is None):
            return None;
        if(self.isEntryModified(name=name) or force or (not os.path.isfile(path))):
            f = open(path, 'wb');
            pickle.dump(self.getEntry(name=name), f, protocol=2);
            f.close();
            self.setEntryModified(name=name, is_modified=False);
        return True;

    def setEntryModified(self, name, is_modified=True):
        self.data[name]['modified']=is_modified;
        if(is_modified):
            self.setModified(is_modified=True);

    def isEntryModified(self, name):
        m=self.data[name].get('modified');
        if(m is not None):
            return m;
        else:
            return True;

    def isModified(self):
        return self.modified;

    def setModified(self, is_modified):
        self.modified=is_modified;

    def save(self, path, force=False):
        """save all entries to one file."""
        if(force or self.isModified()):
            f = open(path, 'wb');
            pickle.dump(self.data, f, protocol=2);
            f.close();
            self.setModified(is_modified=False);
        return True;

    def loadEntry(self, name, path):
        """load one entry from one file."""
        f=open(path, 'rb');
        self.setEntry(name=name, d=pickle.load(f));
        f.close();
        self.setEntryModified(name=name, is_modified=False);
        return True;

    def load(self, path):
        """Load a set of entries all from one file."""
        f=open(path, 'rb');
        newd = pickle.load(f);
        self.data.update(newd);
        f.close();
        return True;

    def getKeyList(self):
        return list(self.data.keys());



from sys import platform
if platform == "linux" or platform == "linux2":
    PLATFORM = 'linux';
elif platform == "darwin":
    PLATFORM = 'osx';
elif platform == "win32":
    PLATFORM = 'windows'

SUPPORTED = False;

INITIAL_DIR ='./';

if(PLATFORM == 'osx'):
    from . import uipath

    def GetFilePath(initial_path=None):
        if(initial_path is None):
            return uipath.uiGetFilePath(initial_path=INITIAL_DIR);
        else:
            return uipath.uiGetFilePath(initial_path);
    def GetDirectory(initial_path=None):
        if(initial_path is None):
            return uipath.uiGetDirectory(initial_path=INITIAL_DIR);
        else:
            return uipath.uiGetDirectory(initial_path);
    def GetSaveFilePath(initial_path=None, file_extension = None):
        if(initial_path is None):
            return uipath.uiGetSaveFilePath(initial_path=INITIAL_DIR, file_extension=file_extension);
        else:
            return uipath.uiGetSaveFilePath(initial_path=initial_path, file_extension=file_extension);

    def Show(path):
        uipath.showInFinder(path);

    def Open(path):
        uipath.openOSX(path);
    SUPPORTED = True;
else:
    def GetFilePath(initial_path=None):
        return None;
    def GetDirectory(initial_path=None):
        return None;
    def GetSaveFilePath(initial_path=None, file_extension = None):
        return None;
    def Show(path):
        return None;
    def Open(path):
        return None;


import os
from visbeat.SourceLocationParser import ParseSourseLocation

class VisBeatExampleVideo(object):
    def __init__(self, name, url, start_beat=None, end_beat = None, display_name = None, code=None, leadin=None, **kwargs):
        self.name = name;
        self.url = url;
        self.start_beat = start_beat;
        self.end_beat = end_beat;
        self._display_name = display_name;
        self._code = code;
        self.leadin = leadin;
        for k in kwargs:
            setattr(self, k, kwargs[k]);
        if(code is None):
            sloc = ParseSourseLocation(self.url);
            self._code = sloc.code;
    # <editor-fold desc="Property: 'code'">
    @property
    def code(self):
        return self._getCode();
    def _getCode(self):
        return self._code;
    @code.setter
    def code(self, value):
        self._setCode(value);
    def _setCode(self, value):
        self._code = value;
    # </editor-fold>

    # def _ytEmbedCode(self):

    # <editor-fold desc="Property: 'display_name'">
    @property
    def display_name(self):
        return self._getDisplayName();
    def _getDisplayName(self):
        if(self._display_name is None):
            return self.name;
        else:
            return self._display_name;
    # </editor-fold>


    def _ytWatchURL(self):
        return 'https://www.youtube.com/watch?v={}'.format(self.code);

    def _ytEmbedURL(self, autoplay=None):
        s = 'https://www.youtube.com/embed/{}'.format(self.code);
        if(autoplay):
            s = s+'?autoplay=1';
        return s;

    def _ytThumbURL(self):
        return 'https://ytimg.googleusercontent.com/vi/{}/default.jpg'.format(self.code);

    def _fancyBoxCode(self, with_label = None):
        """"""
        s = HTMLCode();
        if(with_label):
            s.addLine("<figure>")
        s.addLine('<a class="vb_youtube_fancybox" data-fancybox href="{watchurl}"><img alt="{alt_name}" src="{thumburl}" /></a>'.format(watchurl=self._ytWatchURL(),
                                                                                                                                        alt_name = self.display_name,
                                                                                                                                        thumburl = self._ytThumbURL()))

        if(with_label):
            s.addLine('<figcaption>');
            # s.addLine('<a href="{watchurl}">{displayname}</a>'.format(self.url, self.display_name));
            s.addLine('{displayname}'.format(displayname=self.display_name));
            s.addLine('</figcaption>')
            s.addLine('</figure>');
        return s.string;


from bs4 import BeautifulSoup
class HTMLCode(object):
    def __init__(self, start_string=None):
        self._code = "";
        self._lines = [];
        if(start_string is not None):
            self.add(start_string)
        self._soup = None;

    # <editor-fold desc="Property: 'string'">
    @property
    def string(self):
        return self._getString();
    def _getString(self):
        return BeautifulSoup(self.code).prettify();

    # <editor-fold desc="Property: 'code'">
    @property
    def code(self):
        return self._getCode();
    def _getCode(self):
        return self._code;
    @code.setter
    def code(self, value):
        self._setCode(value);
    def _setCode(self, value):
        self._code = value;
    # </editor-fold>


    # <editor-fold desc="Property: 'soup'">
    @property
    def soup(self):
        return self._getSoup();
    def _getSoup(self):
        if(self._soup is None):
            self._makeSoup();
        return self._soup;

    def _makeSoup(self):
        self._soup = BeautifulSoup(self.code, 'html.parser');
    # @soup.setter
    # def soup(self, value):
    #     self._setSoup(value);
    # def _setSoup(self, value):
    #     self._soup = value;
    # </editor-fold>

    def add(self, s):
        self.code = self.code+s;

    def addLine(self, s):
        self.code = self.code+'\n'+s;

    def startTable(self, id=None, class_ = None, **kwargs):
        targs = dict(table_width="80%", border=1, cellspacing=1, cellpadding=1)
        targs.update(kwargs);
        s = '<table ';
        if(id is not None):
            s = s+'id="{}" '.format(id);
        if(class_ is not None):
            s = s + 'class="{}" '.format(class_);
        s = s+'width="{table_width}" border="{border}" cellspacing="{cellspacing}" cellpadding="{cellpadding}">'.format(**targs);
        self.addLine(s);
        self.addLine('<tbody>');

    def endTable(self):
        self.addLine("</tbody>");
        self.addLine("</table>");

    def startRow(self):
        self.addLine('<tr>');

    def endRow(self):
        self.addLine('</tr>');

    def addColumnLabel(self, label):
        self.addLine('<th scope="col">')
        self.addLine(label);
        self.addLine('</th>');

    def addRowLabel(self, label):
        self.addLine('<th scope="row">')
        self.addLine(label);
        self.addLine('</th>');

    def addRowCell(self, content):
        self.addLine("<td>");
        self.addLine("<div class = 'rowcelldiv'>")
        self.addLine(content);
        self.addLine("</div>")
        self.addLine("</td>");



from .VisBeatExampleVideo import *
from ._music_examples import *

dances.append(VisBeatExampleVideo(display_name = 'Turtle', name='turtle', url='https://www.youtube.com/watch?v=PWD4gktEUAY', start_beat=0));
dances.append(VisBeatExampleVideo(display_name = 'Louie The Cat', name='louie_the_cat', url='https://www.youtube.com/watch?v=hqFG6d86ygI', start_beat=None));



accidental_dances = [];
accidental_dances.append(VisBeatExampleVideo(display_name="Trump IACP Full", name="trump_iacp_full", url='https://www.youtube.com/watch?v=pu4dPTi6SEg', target_n_beats = 7))
accidental_dances.append(VisBeatExampleVideo(display_name="Trump Tax Reform WH.gov Full", name="trump_taxreform_remarks_1", url="https://www.youtube.com/watch?v=zB-lhTEQdKY"))
# accidental_dances.append(VisBeatExampleVideo(display_name="", name="", url=""))


synth = [];
synth.append(VisBeatExampleVideo(name='synth_ball_metronome_w_noise', url = 'https://youtu.be/-mqJIGdro6A', start_beat = None));
synth.append(VisBeatExampleVideo(name='synth_test_ball_1', url = 'https://youtu.be/SjXsMYBJEF8', start_beat = None));
synth.append(VisBeatExampleVideo(name='synth_test_ball_1_gt_impact_sounds', url = 'https://youtu.be/tsdsQm4iTNg', start_beat = None));







from .VisBeatExampleVideo import *
from ._music_examples import *

# dances = [];
dances.append(VisBeatExampleVideo(display_name = 'SNSD: HAHAHA', name='snsd_hahaha', url='https://www.youtube.com/watch?v=OxNsx0eMsGU', start_beat=42));
dances.append(VisBeatExampleVideo(display_name = 'Boston Dynamics: Spot the Dancing Robot', name='spot_the_dancing_robot', url='https://www.youtube.com/watch?v=kHBcVlqpvZ8', start_beat = 9));
dances.append(VisBeatExampleVideo(display_name = 'Michael Jackson: Thriller', name = 'thriller', url = 'https://www.youtube.com/watch?v=wlHUJbl-t7o', start_beat = 4))
dances.append(VisBeatExampleVideo(display_name = 'LMFAO: Party Rock Anthem',name='party_rock', url='https://www.youtube.com/watch?v=WZRMpC_wh0M', start_beat=4, leadin=4))
dances.append(VisBeatExampleVideo(display_name = 'BTS: Boy in Luv', name='boy_in_luv', url='https://www.youtube.com/watch?v=gqz6Adx63w8', start_beat=34));
dances.append(VisBeatExampleVideo(display_name = 'Thays Monaro: Redneck Woman',name='gwilson_redneck_woman', url='https://www.youtube.com/watch?v=1apy8YPygKo', start_beat=65))
dances.append(VisBeatExampleVideo(display_name = 'Music Express: Dancing in the Street', name='dance_cc_dancing_in_the_street', url='https://www.youtube.com/watch?v=fzZ2CoL3IMQ', start_beat=3));
dances.append(VisBeatExampleVideo(display_name = 'Barney and Friends: Mr. Golden Sun',name='barney_mr_golden_sun', url='https://www.youtube.com/watch?v=ya4yyg9XiI4', start_beat=24))
dances.append(VisBeatExampleVideo(display_name = 'M. Express: Friends Forever', name='dance_cc_friends_forever', url='https://www.youtube.com/watch?v=T6ZX6wkJ5Ns', start_beat=2));
dances.append(VisBeatExampleVideo(display_name = 'Just Dance Kids: Gummy Bear',name='jdk_gummy_bear', url='https://www.youtube.com/watch?v=KVE-T2_vLpY', start_beat=60))
dances.append(VisBeatExampleVideo(display_name = 'M. Express: Feelin Good',name='dance_cc_feelin_good', url='https://www.youtube.com/watch?v=uBWp9rr6w08', start_beat=49));
# dances.append(VisBeatExampleVideo(display_name = 'Kiesza: Hideaway', name='hideaway', url='https://www.youtube.com/watch?v=Vnoz5uBEWOA', start_beat=124))
dances.append(VisBeatExampleVideo(display_name = 'SNSD: Genie', name='snsd_genie', url='https://www.youtube.com/watch?v=6SwiSpudKWI', start_beat=141));
dances.append(VisBeatExampleVideo(display_name = 'Momoland Boom Boom', name='momoland_boom_boom', url='https://www.youtube.com/watch?v=0HKfjsM2hSw', start_beat = 145));#33#
dances.append(VisBeatExampleVideo(display_name = 'Snap: Rhythm is a Dancer', name='rhythm_is_a_dancer', url='https://www.youtube.com/watch?v=fDWFVI8PQOI', start_beat=65));
dances.append(VisBeatExampleVideo(display_name = 'Janet Jackson: Rhythm Nation', name='rhythm_nation', url='https://www.youtube.com/watch?v=OAwaNWGLM0c', start_beat = 226));
dances.append(VisBeatExampleVideo(display_name = 'BTS Just One Day', name='bts_just_one_day', url='https://www.youtube.com/watch?v=KQeAg45p2UI', start_beat=200));


# dances.append(VisBeatExampleVideo(display_name = '', name='', url='', start_beat = ));
# dances.append(VisBeatExampleVideo(display_name = '', name='', url='', start_beat = ));
# dances.append(VisBeatExampleVideo(display_name = '', name='', url='', start_beat = ));
# dances.append(VisBeatExampleVideo(display_name = '', name='', url='', start_beat = ));
# dances.append(VisBeatExampleVideo(display_name = '', name='', url='', start_beat = ));







from .TimeSignal import *
import math

class TimeSignal1D(TimeSignal):
    """TimeSignal1D (class): A time signal, and a bunch of convenience functions to go with it.
        Attributes:
            sampling_rate: the sampling rate
            x:  the time signal
    """

    def VBOJECT_TYPE(self):
        return 'TimeSignal1D';

    def __init__(self, path=None, sampling_rate = None, x=None):
        TimeSignal.__init__(self, path=path, sampling_rate=sampling_rate);
        if(x is not None):
            self.x = x;
        #self.initializeBlank(); # will be called by parent

    def initializeBlank(self):
        TimeSignal.initializeBlank(self);#YES KEEP call through weird loops
        self.x = None;

    def getSignal(self, resampled=False):
        return self.x;

    def getSignalSegment(self, time_range):
        signal = self.getSignal();
        seg_start = int(time_range[0]*self.sampling_rate);
        seg_end = int(time_range[1]*self.sampling_rate);
        return signal[seg_start:seg_end];

    def getFullSignal(self):
        return self.x;

    def getSampleAtTime(self, f):
        prev_sample = self.x[int(math.floor(f))];
        next_sample = self.x[int(math.ceil(f))];
        sample_progress = f-np.floor(f);
        return (next_sample*sample_progress)+(prev_sample*(1.0-sample_progress));

    def getSampleAtIndex(self, i):
        return self.x[i];

    def getDuration(self):
        return truediv(len(self.getSignal()), self.sampling_rate);

    def setValueRange(self, value_range=None):
        if(value_range is None):
            value_range = [0,1];
        data = self.x[:];
        currentscale = np.max(data)-np.min(data);
        data = (data/currentscale)*(value_range[1]-value_range[0]);
        data = data-np.min(data)+value_range[0]
        self.x = data;

    def setMaxAbsValue(self, max_abs_val=1.0):
        data = self.x[:];
        currentscale = np.max(np.fabs(data));
        data = (data/currentscale)*max_abs_val;
        self.x = data;



"""
This code was a last minute hack. It works fine enough for parsing youtube urls, but I would use it for any kind of reference.
Based loosely on video backends from django embed video.
"""

import re
import requests
import os
import sys
if sys.version_info.major == 3:
    import urllib.parse as urlparse
else:
    import urllib.parse

class SourceURL(object):
    """
    Modified from django embed video
    """


    allow_https = True
    is_secure = False

    @classmethod
    def SourceLocationType(cls):
        return cls.__name__;

    def __init__(self, source_location):
        self.source_location_type = self.SourceLocationType()
        self._source_location = source_location

    # <editor-fold desc="Property: 'code'">
    @property
    def code(self):
        """
        unique string to identify file
        :return:
        """
        return self._getCode();

    def _getCode(self):
        raise NotImplementedError;

    @code.setter
    def code(self, value):
        self._setCode(value);
    def _setCode(self, value):
        raise NotImplementedError;
    # </editor-fold>


    @property
    def url(self):
        """
        URL of video.
        """
        return self.get_url()
    def get_url(self):
        raise NotImplementedError

    @property
    def protocol(self):
        """
        Protocol used to generate URL.
        """
        return 'https' if self.allow_https and self.is_secure else 'http'

    @property
    def thumbnail(self):
        """
        URL of video thumbnail.
        """
        return self.get_thumbnail_url()

    @classmethod
    def is_valid(cls, url):
        return True if cls.re_detect.match(url) else False

class WebSourceException(Exception):
    """ Parental class for all embed_media exceptions """
    pass


class VideoDoesntExistException(WebSourceException):
    """ Exception thrown if video doesn't exist """
    pass


class UnknownBackendException(WebSourceException):
    """ Exception thrown if video backend is not recognized. """
    pass


class UnknownIdException(VideoDoesntExistException):
    """
    Exception thrown if backend is detected, but video ID cannot be parsed.
    """
    pass


class YoutubeURL(SourceURL):
    """
    for YouTube URLs.
    """

    @classmethod
    def SourceLocationType(cls):
        return 'youtube';

    # Compiled regex (:py:func:`re.compile`) to search code in URL.
    # Example: ``re.compile(r'myvideo\.com/\?code=(?P<code>\w+)')``
    re_code = None

    # Compilede regec (:py:func:`re.compile`) to detect, if input URL is valid for current backend.
    # Example: ``re.compile(r'^http://myvideo\.com/.*')``
    re_detect = None

    # Pattern in which the code is inserted.
    # Example: ``http://myvideo.com?code=%s``
    # :type: str
    pattern_url = None

    pattern_thumbnail_url = None

    re_detect = re.compile(
        r'^(http(s)?://)?(www\.|m\.)?youtu(\.?)be(\.com)?/.*', re.I
    )

    re_code = re.compile(
        r'''youtu(\.?)be(\.com)?/  # match youtube's domains
            (\#/)? # for mobile urls
            (embed/)?  # match the embed url syntax
            (v/)?
            (watch\?v=)?  # match the youtube page url
            (ytscreeningroom\?v=)?
            (feeds/api/videos/)?
            (user\S*[^\w\-\s])?
            (?P<code>[\w\-]{11})[a-z0-9;:@?&%=+/\$_.-]*  # match and extract
        ''',
        re.I | re.X
    )

    pattern_url = '{protocol}://www.youtube.com/embed/{code}'
    pattern_thumbnail_url = '{protocol}://img.youtube.com/vi/{code}/{resolution}'

    resolutions = [
        'maxresdefault.jpg',
        'sddefault.jpg',
        'hqdefault.jpg',
        'mqdefault.jpg',
    ]


    def get_url(self):
        """
        Returns URL folded from :py:data:`pattern_url` and parsed code.
        """
        url = self.pattern_url.format(code=self.code, protocol=self.protocol)
        url += '?' + self.query.urlencode() if self.query else ''
        return url

    def get_thumbnail_url(self):
        """
        Returns thumbnail URL folded from :py:data:`pattern_thumbnail_url` and
        parsed code.

        :rtype: str
        """
        return self.pattern_thumbnail_url.format(code=self.code,
                                                 protocol=self.protocol)

    def _getCode(self):

        match = self.re_code.search(self._source_location)
        if match:
            return match.group('code')

        parsed_url = urllib.parse.urlparse(self._source_location)
        parsed_qs = urllib.parse.parse_qs(parsed_url.query)

        if 'v' in parsed_qs:
            code = parsed_qs['v'][0]
        elif 'video_id' in parsed_qs:
            code = parsed_qs['video_id'][0]
        else:
            raise UnknownIdException('Cannot get ID from `{0}`'.format(self._source_location))

        return code

    def get_thumbnail_url(self):
        """
        Returns thumbnail URL folded from :py:data:`pattern_thumbnail_url` and
        parsed code.

        :rtype: str
        """
        for resolution in self.resolutions:
            temp_thumbnail_url = self.pattern_thumbnail_url.format(
                code=self.code, protocol=self.protocol, resolution=resolution)
            if int(requests.head(temp_thumbnail_url).status_code) < 400:
                return temp_thumbnail_url
        return None



class FilePathURL(SourceURL):
    @classmethod
    def SourceLocationType(cls):
        return 'file_path';

    def __init__(self, source_location):
        self.source_location_type = self.SourceLocationType()
        self._source_location = source_location;

    @classmethod
    def is_valid(cls, source_location):
        return os.path.isfile(source_location);

    def _getCode(self):
        name_parts = os.path.splitext(os.path.basename(self._source_location));
        return name_parts[0];



SOURCE_LOCATION_TYPES = (
    YoutubeURL,
    FilePathURL,
)

def ParseSourseLocation(url):
    """

    :param url:
    :return:
    """
    for backend in SOURCE_LOCATION_TYPES:
        if backend.is_valid(url):
            return backend(url)

    raise UnknownBackendException


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
  name="visbeat3",
  version="0.0.5",
  author="Haofan Wang",
  author_email="haofanwang.ai@gmail.com",
  description="Python3 Implementation for 'Visual Rhythm and Beat' SIGGRAPH 2018",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/haofanwang/visbeat3",
  packages=setuptools.find_packages(),
  install_requires=[
        'numpy',
        'scipy',
        'bs4',
        'librosa',
        'imageio',
        'requests',
        'moviepy',
        'termcolor',
        'youtube-dl',
        'matplotlib',
    ],
  scripts=['bin/dancefer'],
  package_data={'data': ['visbeat3/_assets/*']},
  classifiers=[
  "Programming Language :: Python :: 3",
  "License :: OSI Approved :: Apache Software License",
  "Operating System :: OS Independent",
  ],
)



#Image_Base
#import numpy as np
#import scipy as sp
from PIL import Image as PIM
from PIL import ImageDraw, ImageFont

from .VisBeatImports import *
# from PlotUtils import *


VBMARK_SIZE = 0.135;
VBMMARGIN = 0.02;

def imshow(imdata, new_figure = True):
    if(ISNOTEBOOK):
        #plt.imshow(self.data*0.0039215686274509);#divided by 255
        if(new_figure):
            plt.figure();
        if(len(imdata.shape)<3):
            plt.imshow(imdata, cmap='gray');
        else:
            plt.imshow(imdata);#divided by 255
        plt.axis('off');


class Image(AObject):
    """Image
    """
    IMAGE_TEMP_DIR = None;
    USING_OPENCV=False;
    _VBMRK = None;

    def AOBJECT_TYPE(self):
        return 'Image';

    def __init__(self, data=None, path = None):
        AObject.__init__(self, path=path);
        self.data = data;
        if(path and (not data)):
            self.loadImageData();

    def initializeBlank(self):
        AObject.initializeBlank(self);
        self.data = None;

    def setBlank(self, shape=None):
        if(not shape):
            self.data = np.zeros(self.data.shape);
        else:
            self.data = np.zeros([shape[0], shape[1], ]);

    def loadImageData(self, path = None, force_reload=True):
        if(path):
            self.setPath(path);
        if(self.a_info.get('file_path')):
            if(force_reload or (not self.data)):
                pim = PIM.open(fp=self.a_info['file_path']);
                self.data = np.array(pim);

    @property
    def shape(self):
        return self._getShape();

    def _getShape(self):
        return np.asarray(self.data.shape)[:];

    @property
    def dtype(self):
        return self.data.dtype;

    @property
    def width(self):
        return self._getWidth();

    def _getWidth(self):
        return self.shape[1];

    @property
    def height(self):
        return self._getHeight();

    def _getHeight(self):
        return self.shape[0];

    @property
    def _is_float(self):
        return (self.dtype.kind in 'f');

    @property
    def _is_int(self):
        return (self.dtype.kind in 'iu');

    @property
    def _pixels_float(self):
        if (self._is_float):
            return self.data;
        else:
            return self.data.astype(np.float) * np.true_divide(1.0, 255.0);

    @property
    def _pixels_uint(self):
        if (self._is_int):
            return self.data;
        else:
            return (self.data * 255).astype(np.uint8);

    @property
    def n_channels(self):
        return self._getNChannels();

    def _getNChannels(self):
        if(len(self.data.shape)<3):
            return 1;
        else:
            return self.data.shape[2];

    @staticmethod
    def FromGrayScale(gray_data, color_map = None, format=None, **kwargs):
        if(not color_map):
            newim = Image(data=np.dstack((gray_data,gray_data,gray_data)));
        else:
            cmap=matplotlib.cm.get_cmap(name=color_map);
            #norm = mpl.colors.Normalize(vmin=-20, vmax=10)
            #cmapf = cm.ScalarMappable(norm=norm, cmap=cmap)
            cmapf = matplotlib.cm.ScalarMappable(cmap=cmap);
            if(format is None):
                newim = Image(data=cmapf.to_rgba(gray_data)*255);
            else:
                assert(False), "unrecognized format {}".format(format);

        newim.a_info['ColorType']='FromGrayScale';
        return newim;

    @staticmethod
    def printTempDir():
        print((Image.IMAGE_TEMP_DIR));

    def scaleToValueRange(self, value_range=None):
        is_int = self._is_int;
        if(value_range is None):
            value_range = [0.0,255.0];
        data = self.data.ravel();
        currentmin=np.min(data);
        currentmax=np.max(data);
        currentscale = currentmax-currentmin;
        if(currentscale==0):
            VBWARN("CANNOT SCALE CONSTANT IMAGE")
            return;
        divscale = truediv(1.0,currentscale);
        newrange=(value_range[1]-value_range[0]);
        self.data = (self.data*divscale)*newrange;
        newmin = (currentmin*divscale)*newrange;
        self.data = self.data-newmin+value_range[0]
        if(is_int and value_range[1]>254):
            self.data = self.data.astype(np.int);


    def nChannels(self):
        if(len(self.data.shape)<3):
            return 1;
        else:
            return self.data.shape[2];

    def getClone(self):
        rimg = Image(path=self.a_info.get('file_path'), data=self.data.copy());
        return rimg;

    def getGridPixel(self,x,y,repeatEdge=0):
        xo = x;
        yo = y;

        blackedge = np.zeros(1);
        if(len(self.data.shape)==3):
            blackedge = np.zeros(self.data.shape[2])

        if(y>=self.data.shape[0]):
            if(repeatEdge==1):
                yo = self.data.shape[0]-1
            else:
                return blackedge

        if(y<0):
            if(repeatEdge==1):
                yo = 0
            else:
                return blackedge

        if(x>=self.data.shape[1]):
            if(repeatEdge==1):
                xo=self.data.shape[1]-1
            else:
                return blackedge

        if(x<0):
            if(repeatEdge==1):
                xo = 0
            else:
                return blackedge

        return self.data[yo,xo]

    def getPixel(self, x, y, repeatEdge=0):
        if(isinstance(y,int) and isinstance(x,int)):
            return self.getGridPixel(x,y)
        else:
            yf = int(np.floor(y))
            yc = int(np.ceil(y))
            xf = int(np.floor(x))
            xc = int(np.ceil(x))

            print('getting here?')
            print(xf);
            print(yf);

            tl = self.getGridPixel(xf,yf,repeatEdge)
            tr = self.getGridPixel(xc,yf,repeatEdge)
            bl = self.getGridPixel(xf,yc,repeatEdge)
            br = self.getGridPixel(xc,yc,repeatEdge)

            yalpha = y-yf
            xalpha = x-xf

            topL = tr*xalpha+tl*(1.0-xalpha)
            botL = br*xalpha+bl*(1.0-xalpha)

            retv = botL*yalpha+topL*(1.0-yalpha)
            return retv

    def getShape(self):
        return np.asarray(self.data.shape)[:];

    def getScaled(self, shape=None, shape_xy=None):
        shapeis = (shape is not None);
        shapexyis = (shape_xy is not None);
        assert((shapeis or shapexyis) and not (shapeis and shapexyis)), "Must provide only one of shape or shape_xy for Image.getScaled"
        if(shapeis):
            sz=[shape[0], shape[1], self.data.shape[2]];
        else:
            sz=[shape_xy[1],shape_xy[0],self.data.shape[2]];
        imK = sp.misc.imresize(self.data, size=sz);
        return Image(data=imK);

    def getRotated(self, theta):
        imR = sp.misc.imrotate(self.data, theta);
        return Image(data=imR);


    # def splatAtPixCoord(self, im, location=[0,0]):
    #     self.data[location[0]:(location[0]+im.data.shape[0]), location[1]:(location[1]+im.data.shape[1]),:]=im.data;
    def _splatAtPixCoord(self, im, location=[0,0], **kwargs):
        is_int = self._is_int;
        selftype = self.dtype;
        region0 = [location[0], (location[0]+im.data.shape[0])];
        region1 = [location[1], (location[1]+im.data.shape[1])];
        if(im.n_channels<4):
            self.data[region0[0]:region0[1], region1[0]:region1[1],:]=im.data;
            return;
        if(im.n_channels == 4):
            alphamap = np.moveaxis(np.tile(im._pixels_float[:, :, 3], (3, 1, 1)), [0], [2]);

            blenda = (im._pixels_float[:, :, :3]) * alphamap + self._pixels_float[region0[0]:region0[1], region1[0]:region1[1],:]*(1.0 - alphamap);
            if(is_int):
                blenda = (blenda*255).astype(selftype);
            self.data[region0[0]:region0[1], region1[0]:region1[1], :] = blenda;



    def reflectY(self):
        self.data[:,:,:]=self.data[::-1,:,:];

    def reflectX(self):
        self.data[:,:,:]=self.data[:,::-1,:];

    def PIL(self):
        return PIM.fromarray(np.uint8(self.data));

    def getRGBData(self):
        return self.data[:,:,0:3];

    def normalize(self, scale=1.0):
        self.data = self.data/np.max(self.data.ravel());
        self.data = self.data*scale;

    def show(self, new_figure = True):
        if(ISNOTEBOOK):
            if(new_figure):
                plt.figure();
            if(self.nChannels()==1):
                plt.imshow(self.PIL(), cmap='gray');
            else:
                plt.imshow(self.PIL());#divided by 255
            plt.axis('off');
        else:
            self.PIL().show();

    def writeToFile(self, out_path):
        self.PIL().save(out_path);

    def getEncodedBase64(self):
        return base64.b64encode(self.data);

    def getDataAsString(self):
        return self.data.tostring();

    @staticmethod
    def FromBase64(encoded_data, shape):
        d = base64.decodestring(encoded_data);
        npar = np.frombuffer(d, dtype=np.float64);
        rIm = Image(data=np.reshape(npar, shape));
        return rIm;




    @staticmethod
    def FromDataString(data_string, shape, dtype=None):
        if(dtype is None):
            dtype=np.float64;
        img_1d = np.fromstring(data_string, dtype=dtype);
        reconstructed_img = img_1d.reshape((height, width, -1))


###########################adapted from https://gist.github.com/turicas/1455973##########################
    def _get_font_size(self, text, font_path, max_width=None, max_height=None):
        if max_width is None and max_height is None:
            raise ValueError('You need to pass max_width or max_height')
        font_size = 1
        text_size = self.get_text_size(font_path, font_size, text)
        if (max_width is not None and text_size[0] > max_width) or (max_height is not None and text_size[1] > max_height):
            raise ValueError("Text can't be filled in only (%dpx, %dpx)" % text_size)
        while True:
            if (max_width is not None and text_size[0] >= max_width) or (max_height is not None and text_size[1] >= max_height):
                return font_size - 1
            font_size += 1
            text_size = self.get_text_size(font_path, font_size, text)

    def writeOutlinedText(self, xy, text,
                   font_size=11,
                   max_width=None,
                   max_height=None,
                   encoding='utf8', draw_context = None):
        self.writeText(xy=[xy[0]+3, xy[1]+3], text=text, font_filename='RobotoCondensed-Regular.ttf',
                       font_size=font_size,
                       max_width=max_width,
                       color=(0, 0, 0),
                       max_height=max_height,
                       encoding=encoding, draw_context=draw_context);
        self.writeText(xy=xy, text=text, font_filename='RobotoCondensed-Regular.ttf',
                       font_size=font_size,
                       max_width=max_width,
                       color=(255, 255, 255),
                       max_height=max_height,
                       encoding=encoding, draw_context=draw_context);


    def writeText(self, xy, text, font_filename='RobotoCondensed-Regular.ttf',
                   font_size=11,
                   color=(0, 0, 0),
                   max_width=None,
                   max_height=None,
                   encoding='utf8', draw_context = None):

        x=xy[0];
        y=xy[1];

        font_paths = find_all_files_with_name_under_path(name=font_filename,
                                                         path=os.path.dirname(os.path.abspath(__file__)));
        font_path = font_paths[0];

        if isinstance(text, str):
            text = text.decode(encoding)
        if font_size == 'fill' and (max_width is not None or max_height is not None):
            font_size = self._get_font_size(text, font_path, max_width,
                                           max_height)
        text_size = self._get_text_size(font_path, font_size, text)


        font = ImageFont.truetype(font=font_path, size=font_size);

        # font = ImageFont.truetype(font_filename, font_size)
        if x == 'center':
            x = (self.data.shape[1] - text_size[0]) / 2
        if y == 'center':
            y = (self.data.shape[0] - text_size[1]) / 2

        if(draw_context is None):
            ipil = self.PIL();
            draw = ImageDraw.Draw(ipil)
            draw.text((x, y), text, font=font, fill=color)
            datashape = self.data.shape;
            self.data = np.array(ipil.getdata());
            self.data.shape = datashape;
        else:
            draw_context.text((x, y), text, font=font, fill=color);
        return text_size

    def _get_text_size(self, font_path, font_size, text):
        font = ImageFont.truetype(font_path, font_size)
        return font.getsize(text)

    @staticmethod
    def _VBMark():
        if(Image._VBMRK is None):
            Image._VBMRK = Image(path=GetVBMarkPath());
        return Image._VBMRK;

    def _vbmark(self):
        self._splatAtPixCoord(**self._vbmarker());

    def _vbmarker(self):
        vbm = Image._VBMark();
        wmfrac = min(VBMARK_SIZE * self.width, VBMARK_SIZE * self.height);
        scaleval = min(1.0, np.true_divide(wmfrac, vbm.width));
        if (scaleval < 1.0):
            vbms = vbm.getScaled(shape=[int(vbm.shape[0] * scaleval), int(vbm.shape[1] * scaleval), vbm.shape[2]]);
        else:
            vbms = vbm.clone();
        margin = min(vbm.width, vbm.height, int(VBMMARGIN * self.width), int(VBMMARGIN * self.height));

        return dict(im=vbms, location=[self.height - vbms.height - margin, self.width - vbms.width - margin]);

    def writeTextBox(self, xy, text, box_width, font_filename='RobotoCondensed-Regular.ttf',
                       font_size=11, color=(0, 0, 0), place='left',
                       justify_last_line=False):

        x = xy[0];
        y = xy[1];

        font_paths = find_all_files_with_name_under_path(name=font_filename,
                                                         path=os.path.dirname(os.path.abspath(__file__)));
        font_path = font_paths[0];

        lines = []
        line = []
        words = text.split()
        for word in words:
            new_line = ' '.join(line + [word])
            size = self._get_text_size(font_path, font_size, new_line)
            text_height = size[1]
            if size[0] <= box_width:
                line.append(word)
            else:
                lines.append(line)
                line = [word]
        if line:
            lines.append(line)
        lines = [' '.join(line) for line in lines if line]
        height = y
        for index, line in enumerate(lines):
            height += text_height
            if place == 'left':
                self.writeText((x, height), line,
                                font_filename,
                                font_size,
                                color)
            elif place == 'right':
                total_size = self._get_text_size(font_path, font_size, line)
                x_left = x + box_width - total_size[0]
                self.writeText((x_left, height), line, font_filename,
                                font_size, color)
            elif place == 'center':
                total_size = self._get_text_size(font_path, font_size, line)
                x_left = int(x + ((box_width - total_size[0]) / 2))
                self.writeText((x_left, height), line,
                                font_filename,
                                font_size, color)
            elif place == 'justify':
                words = line.split()
                if (index == len(lines) - 1 and not justify_last_line) or len(words) == 1:
                    self.writeText((x, height), line,
                                    font_filename,
                                    font_size,
                                    color)
                    continue
                line_without_spaces = ''.join(words)
                total_size = self._get_text_size(font_path, font_size,
                                                line_without_spaces)
                space_width = (box_width - total_size[0]) / (len(words) - 1.0)
                start_x = x
                for word in words[:-1]:
                    self.writeText((start_x, height), word,
                                    font_filename,
                                    font_size, color)
                    word_size = self._get_text_size(font_path, font_size,
                                                   word)
                    start_x += word_size[0] + space_width
                last_word_size = self._get_text_size(font_path, font_size,
                                                    words[-1])
                last_word_x = x + box_width - last_word_size[0]
                self.writeText((last_word_x, height), words[-1],
                                font_filename,
                                font_size, color)
        return (box_width, height - y)
#####################################################

from . import Image_CV
if(Image_CV.USING_OPENCV):
    Image.USING_OPENCV = Image_CV.USING_OPENCV;
    Image.RGB2Gray=Image_CV.RGB2Gray;
    Image.GrayToRGB=Image_CV.Gray2RGB;
    Image.cvGoodFeaturesToTrack=Image_CV.cvGoodFeaturesToTrack;
    Image.withFlow=Image_CV.withFlow;
    cvDenseFlowFarneback = Image_CV.cvDenseFlowFarneback;
    # Image. = Image_CV.
    # Image. = Image_CV.
    # Image. = Image_CV.
    # Image. = Image_CV.
    # Image. = Image_CV.
    ocv = Image_CV.ocv;




#CVFunctions
from .VisBeatImports import *
import numpy as np
import scipy as sp
from PIL import Image as PIM
#from VBObject import *
from . import Image as vbImage
import math

DEFAULT_FLOW_HISTOGRAM_BINS = 8;


USING_OPENCV=True
try:
    import cv2 as ocv
except ImportError:
    USING_OPENCV=False;
    AWARN('OpenCV not installed; not importing Image_CV')


#these are functions that use OpenCV that aren't class functions
if(USING_OPENCV):
    def flow2rgb(flow):
        h, w = flow.shape[:2]
        fx, fy = flow[:,:,0], flow[:,:,1]
        ang = np.arctan2(fy, fx) + np.pi
        v = np.sqrt(fx*fx+fy*fy)
        hsv = np.zeros((h, w, 3), np.uint8)
        hsv[...,0] = ang*(180/np.pi/2)
        hsv[...,1] = 255
        #v=np.minimum(v*4, 255);
        v = np.log(v+1);
        vmax = np.max(v[:]);
        vf=v/vmax;
        v=vf*255;
        hsv[...,2] = v;#np.minimum(v*4, 255)
        rgb = ocv.cvtColor(hsv, ocv.COLOR_HSV2BGR)
        return rgb

    def showFlowHSV(flow, new_figure = True):
        if(ISNOTEBOOK):
            #plt.imshow(self.data*0.0039215686274509);#divided by 255
            if(new_figure):
                plt.figure();
            plt.imshow(PIM.fromarray(flow2rgb(flow)));
            plt.axis('off');
        else:
            self.PIL().show();

    def cornerHarris(im, blockSize=None, ksize=None, k=None):
        if (blockSize is None):
            blockSize = 2;
        if (ksize is None):
            ksize = 3;
        if (k is None):
            k = 0.04;
        return ocv.cornerHarris(im, 2, 3, 0.04);

    def cvDenseFlowFarneback(from_image, to_image, pyr_scale=None, levels=None, winsize=None, iterations=None, poly_n=None, poly_sigma=None, flags=None):
        """Can provide numpy arrays or Image objects as input. Returns the flow from->to."""
        # params for Farneback's method
        from_im=from_image;
        to_im=to_image;
        from_is_ob = isinstance(from_image, vbImage.Image);
        to_is_ob = isinstance(to_image, vbImage.Image)

        if(from_is_ob):
            if(from_image.nChannels()>1):
                from_im=from_image.getClone();
                from_im.RGB2Gray();
            from_im=from_im.data;
        if(to_is_ob):
            if(to_image.nChannels()>1):
                to_im=to_image.getClone();
                to_im.RGB2Gray();
            to_im=to_im.data;



        inputs = dict( flow = None,
                            pyr_scale = pyr_scale,
                            levels = levels,
                            winsize = winsize,
                            iterations = iterations,
                            poly_n = poly_n,
                            poly_sigma = poly_sigma,
                            flags=flags);

        default_flow=None;
        default_pyr_scale=0.5;
        default_levels=int(math.log(float(min(to_im.shape)), 2))-4;
        default_winsize = 15;
        default_iterations=3;
        default_poly_n=5;
        default_poly_sigma=1.25;

        use_params = dict(flow=default_flow,
                          pyr_scale=default_pyr_scale,
                          levels=default_levels,
                          winsize=default_winsize,
                          iterations=default_iterations,
                          poly_n=default_poly_n,
                          poly_sigma=default_poly_sigma,
                          flags=0);

        for key in inputs:
            if(inputs[key] is not None):
                use_params[key]=inputs[key];

        return ocv.calcOpticalFlowFarneback(prev=from_im, next=to_im, **use_params);



# These functions are to add to Image class
if(USING_OPENCV):
    def RGB2Gray(self):
        if(self.nChannels()==1):
            return;
        else:
            self.data = ocv.cvtColor(self.data, ocv.COLOR_RGB2GRAY);

    def Gray2RGB(self):
        if(self.nChannels()==1):
            self.data = ocv.cvtColor(self.data, ocv.COLOR_GRAY2RGB)

    def cvGoodFeaturesToTrack(self, maxCorners=None, qualityLevel=None, minDistance=None, corners=None, mask=None, blockSize=None, useHarrisDetector=None):
        assert(self.data is not None), "must provide image to find features";

        if(self.nChannels()==1):
            gray_image=self.data.copy();
        else:
            gray_image=ocv.cvtColor(self.data, cv2.COLOR_RGB2GRAY);

        argd={};
        if(maxCorners is not None):
            d['maxCorners']=maxCorners;
        if(qualityLevel is not None):
            d['qualityLevel']=qualityLevel;
        if(minDistance is not None):
            d['minDistance']=minDistance;
        if(corners is not None):
            d['corners']=corners;
        if(mask is not None):
            d['mask']=mask;
        if(blockSize is not None):
            d['blockSize']=blockSize;
        if(useHarrisDetector is not None):
            d['useHarrisDetector']=useHarrisDetector;
        return ocv.goodFeaturesToTrack(gray_image, **argd);



    def withFlow(self, flow, step=16):
        clone = self.getClone();
        if(clone.nChannels()<3):
            clone.convertToRGB();
        img = clone.data;
        h, w = img.shape[:2]
        y, x = np.mgrid[step/2:h:step, step/2:w:step].reshape(2,-1).astype(int)
        # flow[y, x] is fx, fy
        fx, fy = flow[y,x].T
        # After vstack, each column: x, y, x+fx, y+fy
        # After transpose: each row: x, y, x+fx, y+fy
        # After reshape: each row: [[x, y], [x+fx, y+fy]]
        lines = np.vstack([x, y, x+fx, y+fy]).T.reshape(-1, 2, 2)
        lines = np.int32(lines + 0.5)
        ocv.polylines(img, lines, 0, (0, 255, 0), thickness=3)
        for (x1, y1), (x2, y2) in lines:
            cv2.circle(img, (x1, y1), 1, (0, 255, 0), -1)
        return clone

    



import os
import json
from .VisBeatImports import *
from .AFuncDict import *
from .AObject import AObject

class VBObject(AObject):
    """VBObject (class): This is a paarent class used to implement common serialization and other functions. There ends
        up being three different dictionaries of data.
        a_info - for small labels and such. Part of AObject.
        a_data - for data to be computed in experiments. I generally use this for things I don't want to automatically
                save out to file.
        features - these are features tied to the results of functions, and manager classes (e.g. VideoSource) will save
                these to disk for future use.

        FEATURE_FUNCS is a dictionary mapping the names of features to their corresponding functions.
    """
    FEATURE_FUNCS={};

    def __init__(self, path=None):
        AObject.__init__(self, path=path);

    def initializeBlank(self):
        AObject.initializeBlank(self);
        self.a_info.update({'VBObjectType': self.VBOBJECT_TYPE()});
        self.features = AFuncDict(owner=self, name='features');
        self.features.functions.update(self.FEATURE_FUNCS);

    def saveFeature(self, name, path):
        """Subclasses can implement version of this that will check members for features if those features arent found here."""
        return self.features.saveEntry(name=name, path=path);

    def saveFeatures(self, path):
        return self.features.save(path=path);

    def loadFeature(self, name, path):
        """Subclasses can implement version of this that will check whether feature is registered before loading."""
        return self.features.loadEntry(name=name, path=path);

    def loadFeatures(self, path):
        return self.features.load(path=path);

    def getFeature(self,name, force_recompute=False, **kwargs):
        """Understood to get the value of a feature. can automatically recompute if feature has registered function."""
        params = kwargs;
        assert (not kwargs.get('params')), "STILL TRYING TO USE PARAMS INSTEAD OF KWARGS. FIX THIS";
        return self.features.getValue(name=name, params=kwargs, force_recompute=force_recompute);

    def getFeatureEntry(self, name, params=None, force_recompute=False):
        return self.features.getEntry(name=name, params=params, force_recompute=force_recompute);

    def getFeatureParams(self, name):
        return self.features.getParams(name=name);

    def setFeature(self, name, value, params=None):
        rval = self.features.setEntry(name=name, d=dict(value=value, params=params));
        self.features.setEntryModified(name=name, is_modified=True);
        return rval;

    def removeFeature(self, name, assert_if_absent=True, set_modified=True):
        self.features.removeEntry(name=name, assert_if_absent=assert_if_absent, set_modified=set_modified);



    def hasFeature(self, name):
        """Just checks to see if it's there."""
        return self.features.hasEntry(name=name);

    def getFeatureFunction(self, feature_name):
        return self.features.getFunction(name=feature_name);

    def getFeaturesList(self):
        return self.features.getKeyList();

    def getFeatureFunctionsList(self):
        return self.features.getFunctionList();

    def clearFeatureFiles(self, features_to_clear=None, **kwargs):
        if(self.clear_feature_files_func):
            self.clear_feature_files_func(self, features_to_clear=features_to_clear, **kwargs);
        else:
            VBWARN("CLEAR FEATURE FILES FUNCTION HAS NOT BEEN PROVIDED FOR {} INSTANCE".format(self.VBOBJECT_TYPE()));


    ########### VIRTUAL FUNCTIONS #############

    def AOBJECT_TYPE(self):
        return 'VBObject';

    def VBOBJECT_TYPE(self):
        return self.AOBJECT_TYPE();


    # ##Example of how these functions should be written ina  subclass, for 'AssetManager' class
    # def VBOBJECT_TYPE(self):
    #     return 'AssetManager';
    #
    # def toDictionary(self):
    #     d = VBObject.toDictionary(self);
    #     #serialize class specific members
    #     return d;
    #
    # def initFromDictionary(self, d):
    #     VBObject.initFromDictionary(self, d);
    #     #do class specific inits with d;
